import numpy as np
import pandas as pd
from typing import Optional, List, Dict
from scipy.optimize import linear_sum_assignment
from sklearn.cluster import DBSCAN
from sklearn.mixture import GaussianMixture


def feature_correspondence(
        feature_data: pd.DataFrame,
        sample_per_group: Dict[int, int],
        include_groups: Optional[List[int]],
        mz_tolerance: float,
        rt_tolerance: float,
        min_fraction: float,
        min_likelihood: float
):
    X = feature_data.loc[:, ["mz", "rt"]].to_numpy()
    # scale rt to use the same tolerance in both dimensions
    X[:, 1] *= mz_tolerance / rt_tolerance

    # min_samples estimation using the number of samples
    if include_groups is None:
        min_samples = round(sum(sample_per_group.values()) * min_fraction)
        include_groups = list(sample_per_group.keys())
    else:
        min_samples = _get_min_sample(
            sample_per_group, include_groups, min_fraction
        )
    # DBSCAN clustering
    max_size = 100000
    cluster = _make_initial_cluster(X, mz_tolerance, min_samples, max_size)
    feature_data["cluster_"] = cluster

    # estimate the number of chemical species per cluster
    features_per_cluster = _estimate_n_species(
        feature_data.loc[:, ["sample_", "cluster_", "group_"]],
        min_fraction,
        sample_per_group,
        include_groups,
    )


    # refine each cluster using GMMs: split clusters in a way such that there
    # is only one chemical species per cluster
    cluster_count = 0
    refined_cluster = -1 * np.ones(feature_data.shape[0], dtype=int)
    for name, group in feature_data.groupby("cluster_"):
        n_ft = features_per_cluster[name]
        if (n_ft > 0) and (name > -1):
            # refine clusters using GMMs
            refined_cluster[group.index] = _process_cluster() + cluster_count
            cluster_count += n_ft
    return features_per_cluster

def _get_min_sample(
    samples_per_group: Dict[int, int],
    include_groups: List[int],
    min_fraction: float
) -> int:
    min_sample = np.inf
    for k, v in samples_per_group.items():
        if k in include_groups:
            tmp = round(v * min_fraction)
            min_sample = min(tmp, min_sample)
    return min_sample


def _make_initial_cluster(
        X: np.ndarray,
        eps: float,
        min_samples: int,
        max_size: int
) -> np.ndarray:
    """
    First estimation of matched features using the DBSCAN algorithm.
    Data is split to reduce memory usage.

    Auxiliary function to feature_correspondence.

    Parameters
    ----------
    X : array
        m/z and rt values for each feature
    eps : float
        Used to build epsilon parameter of DBSCAN
    min_samples : int
        parameter to pass to DBSCAN
    max_size : int
        maximum number of rows in X. If the number of rows is greater than
        this value, the data is processed in chunks to reduce memory usage.
    Returns
    -------
    cluster : Series
        The assigned cluster by DBSCAN

    """
    n_rows = X.shape[0]

    if n_rows > max_size:
        # sort X based on the values of the columns and find positions to
        # split into smaller chunks of data.
        sorted_index = np.lexsort(tuple(X.T))
        revert_sort_ind = np.arange(X.shape[0])[sorted_index]
        X = X[sorted_index]

        # indices to split X based on max_size
        split_index = np.arange(max_size, n_rows, max_size)

        # find split indices candidates
        min_diff_x = np.min(np.diff(X.T), axis=0)
        split_candidates = np.where(min_diff_x > eps)[0]
        close_index = np.searchsorted(split_candidates, split_index)

        close_index[-1] = min(split_candidates.size - 1, close_index[-1])
        split_index = split_candidates[close_index] + 1
        split_index = np.hstack((0, split_index, n_rows))
    else:
        split_index = np.array([0, n_rows])

    # clusterize using DBSCAN on each chunk
    dbscan = DBSCAN(eps=eps, min_samples=min_samples, metric="chebyshev")
    n_chunks = split_index.size - 1
    cluster = np.zeros(X.shape[0], dtype=int)
    cluster_counter = 0
    for k in range(n_chunks):
        start = split_index[k]
        end = split_index[k + 1]
        dbscan.fit(X[start:end, :])
        labels = dbscan.labels_
        n_cluster = (np.unique(labels) >= 0).sum()
        labels[labels >= 0] += cluster_counter
        cluster_counter += n_cluster
        cluster[start:end] = labels

    # revert sort on cluster and X
    if n_rows > max_size:
        cluster = cluster[revert_sort_ind]
    return cluster


def _estimate_n_species(
        df: pd.DataFrame,
        min_dr: float,
        sample_per_group: Dict[int, int],
        include_groups: Optional[List[int]]):

    # gets a dataframe with four columns: group, sample, cluster and number
    # of features in the cluster.
    n_ft_cluster = (
        df.groupby(["group_"])
        .value_counts()
        .reset_index()
    )

    # computes maximum number of species where the detection rate is greater
    # than min_dr
    n_species_per_cluster = (
        n_ft_cluster.groupby("group_")
        .apply(
            _get_n_species_per_group,
            min_dr,
            sample_per_group,
            include_groups
        )
    )
    if isinstance(n_species_per_cluster.index, pd.MultiIndex):
        # unstack MultiIndex into a DataFrame and get the highest value
        # estimation of the number of species for each cluster
        n_species_per_cluster = n_species_per_cluster.unstack(-1).max()
    elif isinstance(n_species_per_cluster, pd.DataFrame):
        # If multiple groups are used a DataFrame is obtained
        n_species_per_cluster = n_species_per_cluster.max()
    else:
        # DataFrame with only one row is converted to a Series
        n_species_per_cluster = n_species_per_cluster.iloc[0]

    n_species_per_cluster = n_species_per_cluster.astype(int).to_dict()
    return n_species_per_cluster


def _get_n_species_per_group(
        df: pd.DataFrame,
        min_dr: float,
        samples_per_group: Dict[int, int],
        include_groups: List[int]
) -> pd.DataFrame:
    if df.name in include_groups:
        n_samples = samples_per_group[df.name]
        res = (
            # pivot table where the values are the number of features
            # contributed by a sample
            df.pivot(index="sample_", columns="cluster_", values=0)
            .fillna(0)
            # count how many times a given number of features was obtained
            .apply(lambda x: x.value_counts())
            .div(n_samples)  # convert values to detection rates
            .fillna(0)
            .iloc[::-1]
            .cumsum()
            .iloc[::-1]
            # these three previous steps are a trick to pass detection rates
            # values from higher to lower values.
            .ge(min_dr)     # find values where above the min_dr
            .apply(lambda x: x.iloc[::-1].idxmax() if x.any() else 0)
            # get the maximum number of species above the min_dr
        )
    else:
        c = df.cluster_.unique()
        res = pd.Series(data=np.zeros(c.size), index=c)
    return res


def _process_cluster(df: pd.DataFrame, min_likelihood: float, n_species: int):
    """
    Process each cluster obtained from DBSCAN. Auxiliary function to
    `feature_correspondence`.

    Parameters
    ----------
    df : DataFrame
        feature_data values for a given cluster
    min_likelihood : float
    n_species: int
        Number of features in the cluster, estimated with
        `estimate_features_per_cluster`.

    Returns
    -------
    subcluster : Series
        The subcluster values.
    """

    X = df.loc[:, ["mz", "rt"]].to_numpy()
    # fit a Gaussian mixture model using the cluster data
    gmm = GaussianMixture(n_components=n_species, covariance_type="diag")
    labels = gmm.fit_predict(X)
    likelihood = gmm.score_samples(X)
    df["cluster"] = labels
    df.loc[likelihood < min_likelihood, "cluster"] = -1
    # ft_per_cluster =



def _make_gmm(ft_data: pd.DataFrame, n_feature: int, cluster_name: str):
    """
    fit a gaussian model and set subcluster names for each feature. Auxiliary
    function to process cluster.

    Parameters
    ----------
    ft_data : DataFrame
        The mz and rt columns of the cluster DataFrame
    n_feature : int
        Number of features estimated in the cluster.
    cluster_name: str

    Returns
    -------
    gmm : GaussianMixtureModel fitted with cluster data
    score: The log-likelihood of each feature.
    subcluster : pd.Series with subcluster labels.
    """
    gmm = GaussianMixture(n_components=n_feature, covariance_type="diag")
    gmm.fit(ft_data.loc[:, ["mz", "rt"]].values)
    # scores = pd.Series(data=gmm.score_samples(ft_data), index=ft_data.index)
    ft_data["score"] = gmm.score_samples(ft_data.loc[:, ["mz", "rt"]].values)

    # get index of features in the cases where the number of features is greater
    # than the number of components in the gmm
    noise_index = (ft_data
                   .groupby("sample")
                   .filter(lambda x: x.shape[0] > n_feature))

    if not noise_index.empty:
        noise_index = (noise_index
                       .groupby("sample")
                       .apply(lambda x: _noise_ind(x, n_feature))
                       .droplevel(0)
                       .index)
    else:
        noise_index = noise_index.index

    noise = pd.Series(data="-1", index=noise_index)

    # if the number of features is equal to the number of components in the
    # gmm, each feature is assigned to a cluster using the Hungarian algorithm
    # on the posterior probabilities on each component
    subcluster = (ft_data.loc[ft_data.index.difference(noise_index)]
                  .groupby("sample")
                  .filter(lambda x: x.shape[0] <= n_feature)
                  .groupby("sample")
                  .apply(lambda x: _get_best_cluster(x, gmm))
                  .droplevel(0)
                  .astype(str))
    subcluster = subcluster.apply(lambda x: str(cluster_name) + "-" + x)
    subcluster = pd.concat([noise, subcluster])
    subcluster = subcluster.sort_index()
    # TODO: add here the case where n_features < n_components

    # subcluster = pd.Series(data=gmm.predict(ft_data), index=ft_data.index,
    #                        dtype=str)
    scores = 1
    return gmm, scores, subcluster


def _noise_ind(x, n):
    """
    search the index of samples that are going to be considered as noise.
    Reduces the number of features from a sample in a cluster until the size is
    equal to n
    """
    ind = x["score"].sort_values().index[:(x.shape[0] - n)]
    return x.loc[ind, :]


def _get_best_cluster(x, gmm):
    """
    Assigns a feature to a cluster the posterior probability to each cluster.
    """
    proba = gmm.predict_proba(x.loc[:, ["mz", "rt"]].values)
    rows, cols = proba.shape
    if rows != cols:
        fill = np.zeros(shape=(cols - rows, cols))
        proba = np.vstack((proba, fill))
    _, best_cluster = linear_sum_assignment(proba)
    best_cluster = best_cluster[:rows]
    best_cluster = pd.Series(data=best_cluster, index=x.index)
    return best_cluster



"""
Functions to parse mzML files, optimized for low-memory usage.

Currently, it retrieves spectrum and chromatogram data.

build_offset_list : Finds offsets with the location of spectra and chromatograms
get_spectrum : Extracts data from a spectrum using offsets.
get_chromatogram : Extracts data from a chromatogram using offsets.

"""

import base64
import numpy as np
import re
import zlib
from os import SEEK_END
from os.path import getsize
from typing import Dict, List, Tuple
from xml.etree.ElementTree import fromstring
from xml.etree.ElementTree import Element

# Rationale for the implementation:
# mzML files can have typical sizes that span from 50 MiB to more than 1000 MiB.
# mzML files are basically xml files with the acquired data and metadata
# associated with the acquisition conditions. The data is inside <spectrum> and
# <chromatogram> elements.  From the expected file sizes, it is clear that
# loading the entire file in memory is not desirable.
# There are tools to parse xml files in an incremental way (such as iterparse
# inside the xml module), but the problem with this approach is that accessing a
# specific scan in the file turns out to be rather slow.
# Indexed mzML files have an <indexedmzML> tag that allows to search in a fast
# way the location of spectrum and chromatogram tags.
# Taking advantage of this information, this module search spectrum and
# chromatogram elements and only loads the part associated with the selected
# spectrum or chromatogram.
# The function build_offset_list creates the list of offsets associated with
# chromatograms and spectra using the information in <indexedmzML>. In the
# case of non-indexed files, the offset list is built from scratch.
# The functions get_spectrum and get_chromatogram takes the offset information
# and extracts the relevant data from each spectrum/chromatogram.


# mzml specification:
# https://www.psidev.info/mzML
# terms defined here:
# https://raw.githubusercontent.com/HUPO-PSI/psi-ms-CV/master/psi-ms.obo
UNSUPPORTED_COMPRESSION = {
    "MS:1002312",
    "MS:1002313",
    "MS:1002314",
    "MS:1002746",
    "MS:1002747",
    "MS:1002748",
    "MS:1003088",
    "MS:1003089",
    "MS:1003090",
}
ZLIB = "MS:1000574"
MZ_ARRAY = "MS:1000514"
INT_ARRAY = "MS:1000515"
TIME_ARRAY = "MS:1000595"
MS_LEVEL = "MS:1000511"
TIME = "MS:1000016"  # time value in scans
NEGATIVE_POLARITY = "MS:1000129"
POSITIVE_POLARITY = "MS:1000130"
# time units
SECONDS = "UO:0000010"
MINUTES = "UO:0000031"

# data types
FLOAT16 = "MS:1000521"
FLOAT32 = "MS:1000518"
FLOAT64 = "MS:1000523"
INT32 = "MS:1000519"
INT64 = "MS:1000522"
DATA_TYPES = {
    FLOAT16: np.float16,
    FLOAT32: np.float32,
    FLOAT64: np.float64,
    INT32: np.int32,
    INT64: np.int64,
}


def build_offset_list(filename: str) -> Tuple[List[int], List[int], int]:
    """
    Finds the offset values in the file where Spectrum or Chromatogram elements
    start.

    Parameters
    ----------
    filename : path to a mzML file

    Returns
    -------
    spectra_offset : list
        Offsets where spectrum element start.
    chromatogram_offset : list
        Offsets where chromatogram elements start.
    index_offset : int
        Offset where the index starts. If the file is not indexed, return the
        file size.

    """
    if is_indexed(filename):
        index_offset = _get_index_offset(filename)
        spectra_offset, chromatogram_offset = _build_offset_list_indexed(
            filename, index_offset
        )
    else:
        spectra_offset, chromatogram_offset = _build_offset_list_non_indexed(
            filename
        )
        index_offset = getsize(filename)
    return spectra_offset, chromatogram_offset, index_offset


def get_spectrum(
    filename: str,
    spectra_offset: List[int],
    chromatogram_offset: List[int],
    index_offset: int,
    n: int,
) -> Dict:
    """
    Extracts m/z, intensity, polarity , time, and ms level from the nth scan.
    Parameters
    ----------
    filename : str
        path to mzML file.
    spectra_offset : list
        Offset list obtained from `_build_offset_list`.
    chromatogram_offset : list
        Offset list obtained from `_build_offset_list`.
    index_offset : int
        Offset obtained from `_build_offset_list`.
    n : int
        scan number to select.

    Returns
    -------
    spectrum : dictionary with spectrum data.

    """
    xml_str = _get_xml_data(
        filename,
        spectra_offset,
        chromatogram_offset,
        index_offset,
        n,
        "spectrum"
    )
    elements = list(fromstring(xml_str))
    spectrum = dict()
    for el in elements:
        tag = el.tag
        if tag == "cvParam":
            accession = el.attrib.get("accession")
            if accession == MS_LEVEL:
                spectrum["ms_level"] = int(el.attrib.get("value"))
            elif accession == NEGATIVE_POLARITY:
                spectrum["polarity"] = -1
            elif accession == POSITIVE_POLARITY:
                spectrum["polarity"] = 1
        elif tag == "scanList":
            spectrum["time"] = _get_time(el)
        elif tag == "binaryDataArrayList":
            sp_data = _parse_binary_data_list(el)
            spectrum.update(sp_data)
    return spectrum


def get_chromatogram(
    filename: str,
    spectra_offset: List[int],
    chromatogram_offset: List[int],
    index_offset: int,
    n: int,
) -> Dict:
    """
    Extracts time and intensity from xml chunk.

    Parameters
    ----------
    filename : path to mzML file
    spectra_offset : offset list obtained from _build_offset_list
    chromatogram_offset : offset list obtained from _build_offset_list
    index_offset : offset obtained from _build_offset_list
    n : chromatogram number

    Returns
    -------
    chromatogram: dict

    """
    xml_str = _get_xml_data(
        filename,
        spectra_offset,
        chromatogram_offset,
        index_offset,
        n,
        "chromatogram"
    )
    elements = fromstring(xml_str)
    name = elements.attrib.get("id")
    chromatogram = dict(name=name)
    for el in list(elements):
        tag = el.tag
        if tag == "binaryDataArrayList":
            chrom_data = _parse_binary_data_list(el)
            chromatogram.update(chrom_data)
    return chromatogram


class ReverseReader:
    """
    Reads file objects starting from the EOF.

    """

    def __init__(self, filename: str, buffer_size: int, **kwargs):
        """
        Constructor object

        Parameters
        ----------
        filename : str,
            Path to the file
        buffer_size : int
            size of the chunk to get when using the `read_chunk` method
        **kwargs :
            keyword arguments to pass to the open function
        """

        self.file = open(filename, **kwargs)
        self.buffer_size = buffer_size
        self._size = self.file.seek(0, SEEK_END)
        self.offset = self._size
        self._is_complete = False

    @property
    def offset(self) -> int:
        return self._offset

    @offset.setter
    def offset(self, value: int):
        if value < 0:
            value = 0
            self._is_complete = True
        self._offset = value
        self.file.seek(value)

    def __enter__(self):
        return self

    def __exit__(self, t, value, traceback):
        self.file.close()

    def read_chunk(self) -> str:
        """
        Reads a chunk of data from the file, starting from the end. If the
        beginning of the file has been reached, it returns None.

        Returns
        -------
        chunk : str
        """
        if self._is_complete:
            res = None
        else:
            self.offset = self.offset - self.buffer_size
            res = self.file.read(self.buffer_size)
        return res

    def reset(self):
        """
        Set the position of the reader to the EOF.
        """
        self.offset = self._size
        self._is_complete = False


def _read_binary_data_array(element: Element) -> Tuple[np.ndarray, str]:
    """
    Extracts the binary data and data kind from a binaryArray element.

    Parameters
    ----------
    element: Element

    Returns
    -------
    data : array
    kind : can be one of {"mz", "spint", "time"}

    """
    has_zlib_compression = False
    data = None
    kind = None
    units = None
    dtype = None
    for e in element:
        tag = e.tag
        if tag == "binary":
            data = e.text
        else:
            accession = e.attrib.get("accession")
            if accession in UNSUPPORTED_COMPRESSION:
                msg = "Currently only zlib compression is supported."
                raise NotImplementedError(msg)
            elif accession == ZLIB:
                has_zlib_compression = True
            elif accession == INT_ARRAY:
                kind = "spint"
            elif accession == MZ_ARRAY:
                kind = "mz"
            elif accession == TIME_ARRAY:
                kind = "time"
                units = e.attrib.get("unitAccession")
            if accession in DATA_TYPES:
                dtype = DATA_TYPES[accession]
    if data:
        data = base64.b64decode(data)
        if has_zlib_compression:
            data = zlib.decompress(data)
        data = np.frombuffer(data, dtype=dtype).copy()
        if kind == "time":
            data = _time_to_seconds(data, units)
    else:
        data = np.array([])
    return data, kind


def _parse_binary_data_list(element: Element) -> Dict[str, np.ndarray]:
    """
    Extracts the data from a binaryDataArrayList.

    Parameters
    ----------
    element: Element

    Returns
    -------
    dictionary that maps data kind to its correspondent data array.

    """
    res = dict()
    for e in element:
        data, data_type = _read_binary_data_array(e)
        res[data_type] = data
    return res


def is_indexed(filename: str) -> bool:
    """
    Checks if a mzML file is indexed.

    Parameters
    ----------
    filename: str

    Returns
    -------
    bool

    Notes
    -----
    This function assumes that the mzML file is validated. It looks up the last
    closing tag, that should be </indexedmzML> if the file is indexed.

    """
    with ReverseReader(filename, 1024, mode="r") as fin:
        end_tag = "</indexedmzML>"
        chunk = fin.read_chunk()
        res = chunk.find(end_tag) != -1
    return res


def _get_index_offset(filename: str) -> int:
    """
    Search the byte offset where the indexListOffset element starts.

    Parameters
    ----------
    filename : str

    Returns
    -------
    index_offset : int

    """
    tag = "<indexListOffset>"
    # reads mzml backwards until the tag is found
    # we exploit the fact that according to the mzML schema the
    # indexListOffset should have neither attributes nor subelements
    with ReverseReader(filename, 1024, mode="r") as fin:
        xml = ""
        ind = -1
        while ind == -1:
            chunk = fin.read_chunk()
            xml = chunk + xml
            ind = chunk.find(tag)
        # starts at the beginning of the text tag
        start = ind + len(tag)
        xml = xml[start:]
        end = xml.find("<")
    index_offset = int(xml[:end])
    return index_offset


def _build_offset_list_non_indexed(
        filename: str
) -> Tuple[List[int], List[int]]:
    """
    Builds manually the indices for non-indexed mzML files.

    Parameters
    ----------
    filename : str

    Returns
    -------

    """
    # indices are build by finding the offset where spectrum or chromatogram
    # elements starts.
    spectrum_regex = re.compile("<spectrum .[^(><.)]+>")
    chromatogram_regex = re.compile("<chromatogram .[^(><.)]+>")
    ind = 0
    spectrum_offset_list = list()
    chromatogram_offset_list = list()
    with open(filename) as fin:
        while True:
            line = fin.readline()
            if line == "":
                break
            spectrum_offset = _find_spectrum_tag_offset(line, spectrum_regex)
            if spectrum_offset:
                spectrum_offset_list.append(ind + spectrum_offset)
            chromatogram_offset = _find_chromatogram_tag_offset(
                line, chromatogram_regex
            )
            if chromatogram_offset:
                chromatogram_offset_list.append(ind + chromatogram_offset)
            ind += len(line)
    return spectrum_offset_list, chromatogram_offset_list


def _find_spectrum_tag_offset(line: str, regex: re.Pattern) -> int:
    if line.lstrip().startswith("<spectrum"):
        match = regex.search(line)
        if match:
            start, end = match.span()
        else:
            start = None
        return start


def _find_chromatogram_tag_offset(line: str, regex: re.Pattern) -> int:
    if line.lstrip().startswith("<chromatogram"):
        match = regex.search(line)
        if match:
            start, end = match.span()
        else:
            start = None
        return start


def _build_offset_list_indexed(
    filename: str,
    index_offset: int
) -> Tuple[List[int], List[int]]:
    """
    Builds a list of offsets where spectra and chromatograms are stored.

    Parameters
    ----------
    filename : str
    index_offset : int
        offset obtained from _get_index_offset

    Returns
    -------
    spectra_offset : List
        offset where spectra are stored
    chromatogram_offset : List
        offset where chromatograms are stored

    """
    end_tag = "</indexList>"
    with open(filename, "r") as fin:
        fin.seek(index_offset)
        index_xml = fin.read()
        end = index_xml.find(end_tag)
        index_xml = index_xml[: end + len(end_tag)]
    index_xml = fromstring(index_xml)

    spectra_offset = list()
    chromatogram_offset = list()
    for index in index_xml:
        for offset in index:
            value = int(offset.text)
            if index.attrib["name"] == "spectrum":
                spectra_offset.append(value)
            elif index.attrib["name"] == "chromatogram":
                chromatogram_offset.append(value)
    return spectra_offset, chromatogram_offset


def _get_xml_data(
    filename: str,
    spectra_offset: List[int],
    chromatogram_offset: List[int],
    index_offset: int,
    n: int,
    kind: str,
) -> str:
    """
    Get the xml string associated with a spectrum or chromatogram.

    Parameters
    ----------
    filename : str
    spectra_offset : list
        offsets obtained from _build_offset_list
    chromatogram_offset : list
        offsets obtained from _build_offset_list
    index_offset : int
        offset obtained from _get_index_offset
    n : int
        number of spectrum/chromatogram to select
    kind : {"spectrum", "chromatogram"}

    Returns
    -------
    str

    """
    if kind == "spectrum":
        l, other = spectra_offset, chromatogram_offset
        end_tag = "</spectrum>"
    elif kind == "chromatogram":
        l, other = chromatogram_offset, spectra_offset
        end_tag = "</chromatogram>"
    else:
        raise ValueError("Kind must be `spectrum` or `chromatogram`")

    start = l[n]
    # Here we search the closest offset to start such that the complete data
    # is contained in the text
    try:
        end = l[n + 1]
    except IndexError:
        try:
            end = other[0]
        except IndexError:
            end = index_offset

    if end < start:
        end = index_offset

    with open(filename, "r") as fin:
        fin.seek(start)
        chunk = fin.read(end - start)
    end = chunk.find(end_tag)
    return chunk[: end + len(end_tag)]


def _get_time(element):
    for e in list(element):
        tag = e.tag
        if tag == "scan":
            for ee in list(e):
                accession = ee.attrib.get("accession")
                if accession == TIME:
                    value = float(ee.attrib.get("value"))
                    units = ee.attrib.get("unitAccession")
                    value = _time_to_seconds(value, units)
                    return value


def _time_to_seconds(value: float, units: str):
    if units == MINUTES:
        value = value * 60
    return value



import pickle
from pathlib import Path
from typing import Optional, Tuple, List, Dict, Union, Callable
from .fileio import MSData
from .lcms import Roi
from .utils import is_notebook
from multiprocessing import Pool
from functools import partial
from time import time_ns
from tqdm import tqdm as cli_bar
from tqdm.notebook import tqdm as notebook_bar
import numpy as np
import pandas as pd

# TODO: testear y refactorear feature_correspondence. Agregar funcionalidad
#   para plotear ROI y features. Incluir funcionalidad para crear datacontainer.
#   agregar funcionalidad para almacenar ROI en disco.
#   la idea seria usar un parametro store_to_disk que se use para crear una
#   carpeta donde se almacenen las ROI.

# TODO: raise error messages if the functions are called in the wrong order
#   This could be done implementing a tag associated with each processing step
#   If the tag changes, the process should be done again. tag can be generated
#   from the hash of the parameters.

# TODO: add options to add new_samples

# TODO: add path variable to load existing assay data

class Assay:

    def __init__(
        self,
        path: Union[str, List[str], Path],
        sample_metadata: Optional[pd.DataFrame] = None,
        name: Optional[str] = None,
        ms_mode: str = "centroid",
        instrument: str = "qtof",
        separation: str = "uplc",
    ):
        # path for mzml files
        path_list = _get_path_list(path)
        name2path = {x.stem: x for x in path_list}
        self._name_to_path = name2path

        # sample metadata
        sample2code = dict(zip(name2path.keys(), range(len(name2path))))
        self._sample_to_code = sample2code
        self._sample_to_group = None
        self._sample_per_group = None
        self._group_factors = None
        self.sample_metadata = sample_metadata

        # assay name and assay storage dir
        self.name = name
        data_dir_name = "{}.tidyms-assay".format(self.name)
        self._data_path = Path(data_dir_name)
        self._roi_path = self._data_path.joinpath("roi")
        self._feature_path = self._data_path.joinpath("feature")
        if not self._data_path.is_dir():
            self._data_path.mkdir()
            self._roi_path.mkdir()
            self._feature_path.mkdir()
        else:
            check = self._feature_path.is_dir() and self._roi_path.is_dir()
            if not check:
                msg = "{} is not a valid assay path.".format(path)
                raise ValueError(msg)

        self.feature_table = None
        self._params = dict()
        ms_data_params = {
            "ms_mode": ms_mode,
            "instrument": instrument,
            "separation": separation
        }
        self._params["MSData"] = ms_data_params

    @property
    def sample_metadata(self) -> pd.DataFrame:
        return self._sample_metadata

    @sample_metadata.setter
    def sample_metadata(self, value: pd.DataFrame):
        # TODO: validate data
        if value is None:
            self._sample_to_group = {x: 0 for x in self._name_to_path}
            self._sample_per_group = {0: len(self._name_to_path)}
            self._group_factors = [0]
        elif "group_" in value:
            groups, groups_factors = pd.factorize(value.group_)
            counts = np.unique(groups, return_counts=True)
            self._sample_to_group = dict(zip(value.index, groups))
            self._sample_per_group = dict(zip(*counts))
            self._group_factors = groups_factors
        self._sample_metadata = value

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: Optional[str]):
        if value is None:
            value = "tidyms-assay-{}".format(time_ns())
        # TODO: check valid file names
        elif not value.isalnum():
            msg = "Value must be a alphanumeric string. Got '{}'".format(value)
            raise ValueError(msg)
        self._name = value

    def make_roi(self, **kwargs):
        """
        Builds regions of interest from raw data.

        ROIs are stored in the `roi` attribute. See the Notes for a detailed
        description of the algorithm used.

        Parameters
        ----------
        tolerance : positive number or None, default=None
            m/z tolerance to connect values across scans. If None, the value is
            set based on the value of ``self.instrument``. If
            ``self.instrument`` is ``"qtof"``, the tolerance is ``0.01``. If
            ``self.instrument`` is ``"orbitrap"``, the tolerance is ``0.005``
        max_missing : non-negative integer or None, default=None
            maximum number of consecutive missing values in a valid  ROI. If
            ``None``, the value is set to ``1``.
        min_length : positive integer or None, default=None
            The minimum length of a valid ROI. If, ``None``, the value is set
            based on ``self.separation``. If ``self.separation`` is ``"uplc"``,
            the value is set to ``10``. If ``self.separation`` is ``"hplc"``,
            the value is set to ``20``.
        min_intensity : non-negative number , default=0.0
            Minimum intensity in a valid ROI.
        pad: int or None, default=None
            Pad dummy values to the left and right of the ROI. This produces
            better peak picking results when searching low intensity peaks in a
            ROI. Using None set the value to ``2`` if ``self.separation`` is
            ``"uplc"`` or ``"hplc"``.
        multiple_match : {"closest", "reduce"}, default="reduce"
            How peaks are matched when there is more than one valid match. If
            ``"closest"`` is used, the closest peak is assigned as a match and
            the others are used to create new ROIs. If ``"reduce"`` is used,
            unique m/z and intensity values are generated using the reduce
            function in `mz_reduce` and `sp_reduce` respectively.
        mz_reduce : "mean" or Callable, default="mean"
            Function used to reduce m/z values. If ``"mean"`` is used, the mean
            value of all valid m/z is used. Any function that accepts numpy
            arrays and return numbers can be used. Used only when
            `multiple_match` is set to ``"reduce"``. See the following
            prototype:

            .. code-block:: python

                def mz_reduce(mz_match: np.ndarray) -> float:
                    pass

        sp_reduce : {"mean", "sum"} or Callable, default="sum"
            Function used to reduce intensity values. ``"mean"`` computes the
            mean intensity and ``"sum"`` computes the total intensity. Any
            function that accepts numpy arrays and return numbers can be used.
            Only used when `multiple_match` is set to ``"reduce"``. See the
            prototype shown on `mz_reduce`.
        targeted_mz : numpy.ndarray or None, default=None
            A list of m/z values to perform a targeted ROI creation. If this
            value is provided, only ROI with these m/z values will be created.
        ms_level : int, default=1
            ms level used to build the ROI.
        start : int, default=0
            Create ROI starting at this scan
        end : int or None, default=None
            Stop ROI creation at this scan. If None, stops after the last scan
            in the file.
        start_time : float, default=0.0
            Use scans starting at this acquisition time.
        end_time : float or None, default=None
            Stops when the acquisition time is higher than this value.
        min_snr : positive number, default=10.0
            Minimum signal-to-noise ratio of the peaks. Used only to convert
            profile data to centroid mode
        min_distance : positive number or None, default=None
            Minimum distance between consecutive peaks. If ``None``, the value
            is set to 0.01 if ``self.instrument`` is ``"qtof"`` or to 0.005 if
            ``self.instrument`` is ``"orbitrap"``. Used only to convert profile
            data to centroid mode.
        n_jobs: int, default=1
            Number of parallel process used to create ROI. If ``None``, use all
            available cores. If there are errors while using multiple cores in
            Jupyter Notebooks, the kernel will restart and no error message
            will be displayed.
        verbose : bool, default=True
            Display a message showing progress.

        Notes
        -----
        ROIs are built using the method described in [TR08]_ with slight
        modifications.

        A ROI is modelled as a combination of three arrays with the same size:
        m/z, intensity and time. ROIs are created and extended connecting
        m/z values across successive scans using the following method:

        1.  The m/z values in The first scan are used to initialize a list of
            ROI. If `targeted_mz` is used, the ROI are initialized using this
            list.
        2.  m/z values from the next scan extend the ROIs if they are closer
            than `tolerance` to the mean m/z of the ROI. Values that don't match
            any ROI are used to create new ROIs and are appended to the ROI
            list (only if `targeted_mz` is not used).
        3.  If more than one m/z value is within the tolerance threshold,
            m/z and intensity values are computed according to the
            `multiple_match` strategy.
        4.  If a ROI can't be extended with any m/z value from the new scan,
            it is extended using NaNs.
        5.  If the last `max_missing` values of a ROI are NaN, the ROI is
            flagged as finished. If the maximum intensity of a finished ROI
            is greater than `min_intensity` and the number of points is greater
            than `min_length`, then the ROI is flagged as valid. Otherwise, the
            ROI is discarded.
        6.  Repeat from step 2 until no more scans are available.

        Returns
        -------
        roi : list[Roi]
            A list with the detected regions of interest.

        Raises
        ------
        ValueError
            If the data is not in centroid mode.

        See Also
        --------
        lcms.Roi : Representation of a ROI.

        References
        ----------
        .. [TR08] Tautenhahn, R., Böttcher, C. & Neumann, S. Highly sensitive
            feature detection for high resolution LC/MS. BMC Bioinformatics 9,
            504 (2008). https://doi.org/10.1186/1471-2105-9-504

        """
        n_samples = len(self._name_to_path)
        try:
            verbose = kwargs.pop("verbose")
        except KeyError:
            verbose = True

        try:
            n_jobs = kwargs.pop("n_jobs")
        except KeyError:
            n_jobs = 1

        func = partial(
            _make_roi_worker,
            ms_data_params=self._params["MSData"],
            make_roi_params=kwargs,
            roi_path=self._roi_path
        )
        bar = _get_progress_bar()
        if n_jobs == 1:
            results = (func(x) for x in self._name_to_path.items())
            if verbose:
                results = bar(results, total=n_samples)
            tuple(results)
        else:
            with Pool(n_jobs) as pool:
                results = pool.imap(func, self._name_to_path.items())
                if verbose:
                    print("Creating ROI in {} samples.".format(n_samples))
                    results = bar(results, total=n_samples)
                tuple(results)
        self._params["make_roi"] = kwargs

    def load_roi(self, name: str) -> List[Roi]:
        file_path = self._roi_path.joinpath(name + ".pickle")
        with file_path.open("rb") as fin:
            roi = pickle.load(fin)
        return roi

    def load_feature(self, name: str) -> pd.DataFrame:
        file_path = self._feature_path.joinpath(name + ".pickle")
        with file_path.open("rb") as fin:
            df = pickle.load(fin)
        return df

    def detect_features(
        self,
        smoothing_strength: Optional[float] = 1.0,
        find_peaks_params: Optional[dict] = None,
        descriptors: Optional[dict] = None,
        filters: Optional[dict] = None,
        verbose: bool = True,
        n_jobs: Optional[int] = 1
    ):
        """
        Perform feature detection on LC-MS centroid samples.

        Parameters
        ----------
        smoothing_strength: positive number, optional
            Width of a gaussian window used to smooth the ROI. If None, no
            smoothing is applied.
        find_peaks_params : dict, optional
            parameters to pass to :py:func:`tidyms.peaks.detect_peaks`
        descriptors : dict, optional
            descriptors to pass to :py:func:`tidyms.peaks.get_peak_descriptors`
        filters : dict, optional
            filters to pass to :py:func:`tidyms.peaks.get_peak_descriptors`
        descriptors : dict, optional
            pass custom descriptors to :py:func:`tidyms.peaks.get_peak_descriptors`
        filters : dict, optional
            pass custom filters to :py:func:`tidyms.peaks.get_peak_descriptors`
        verbose: bool

        Returns
        -------
        feature_table: DataFrame
            A Pandas DataFrame where each row is a feature detected in a sample and
            each column is a feature descriptor. By default the following
            descriptors are computed:

            mz
                weighted average of the m/z in the peak region.
            mz std
                standard deviation of the m/z in the peak region.
            rt
                weighted average of the retention time in the peak region.
            width
                Chromatographic peak width.
            height
                Height of the chromatographic peak minus the baseline.
            area
                Area of the chromatographic peak. minus the baseline area.
            sample
                The sample name where the feature was detected.

            Also, two additional columns have information to search each feature
            in its correspondent Roi:

            roi_index :
                index in the list of ROI where the feature was detected.
            peak_index :
                index of the peaks attribute of each ROI associated to the feature.

        Notes
        -----
        Features are detected as follows:

        1.  Default parameters are set based on the values of the parameters
            `instrument` and `separation`.
        2.  Regions of interest (ROI) are detected in each sample. See the
            documentation of :py:meth:`tidyms.fileio.MSData.make_roi` for a detailed
            description of how ROI are created from raw data.
        3.  Features (chromatographic peaks) are detected on each ROI. See
            :py:meth:`tidyms.lcms.Chromatogram.find_peaks` for a detailed
            description of how peaks are detected and how descriptors are computed.

        """
        func = partial(
            _detect_features_worker,
            smoothing_strength=smoothing_strength,
            descriptors=descriptors,
            filters=filters,
            find_peaks_params=find_peaks_params
        )
        p_iter = (self._roi_path.joinpath(x + ".pickle") for x in self._name_to_path)
        bar = _get_progress_bar()
        n_samples = len(self._name_to_path)
        if n_jobs == 1:
            results = (func(x) for x in p_iter)
            if verbose:
                results = bar(results, total=n_samples)
            tuple(results)
        else:
            with Pool(n_jobs) as pool:
                results = pool.imap(func, p_iter)
                if verbose:
                    msg = "Searching features in {} samples.".format(n_samples)
                    print(msg)
                    results = bar(results, total=n_samples)
                tuple(results)

    def build_feature_table(self):
        # create feature table
        df_list = list()
        for name in self._name_to_path:
            df = self.load_feature(name)
            df["sample_"] = self._sample_to_code[name]
            group = self._sample_to_group.get(name)
            if group is None:
                df["group_"] = 0
            else:
                df["group_"] = group
            df_list.append(df)
        feature_table = pd.concat(df_list).reset_index(drop=True)
        feature_table["roi index"] = feature_table["roi index"].astype(int)
        feature_table["peak index"] = feature_table["peak index"].astype(int)
        feature_table["sample_"] = feature_table["sample_"].astype(int)
        feature_table["group_"] = feature_table["group_"].astype(int)
        save_path = self._data_path.joinpath("feature-table.pickle")
        feature_table.to_pickle(save_path)
        self.feature_table = feature_table


def _get_path_list(path: Union[str, List[str], Path]) -> List[Path]:
    if isinstance(path, str):
        path = Path(path)

    if isinstance(path, list):
        path_list = [Path(x) for x in path]
        for p in path_list:
            # check if all files in the list exists
            if not p.is_file():
                msg = "{} doesn't exist".format(p)
                raise ValueError(msg)
    else:
        if path.is_dir():
            path_list = list(path.glob("*.mzML"))
        elif path.is_file():
            path_list = [path]
        else:
            msg = ("Path must be a string or Path object pointing to a "
                   "directory with mzML files or a list strings with the "
                   "absolute path to mzML")
            raise ValueError(msg)
    return path_list


def _load_roi(file_path: Path) -> List[Roi]:
    with file_path.open("rb") as fin:
        roi = pickle.load(fin)
    return roi


def _save_roi(file_path: Path, roi: List[Roi]):
    with file_path.open("wb") as fin:
        pickle.dump(roi, fin)


def _make_roi_worker(arg, ms_data_params, make_roi_params, roi_path: Path):
    sample_name, sample_path = arg
    ms_data = MSData(sample_path, **ms_data_params)
    roi = ms_data.make_roi(**make_roi_params)
    file_path = roi_path.joinpath(sample_name + ".pickle")
    _save_roi(file_path, roi)


def _get_progress_bar():
    if is_notebook():
        progress_bar = notebook_bar
    else:
        progress_bar = cli_bar
    return progress_bar


def _detect_features_worker(
        file_path: Path,
        smoothing_strength: Optional[float] = 1.0,
        descriptors: Optional[dict] = None,
        filters: Optional[dict] = None,
        find_peaks_params: Optional[dict] = None
) -> pd.DataFrame:
    """
    Builds a DataFrame with feature descriptors. Aux function for
    detect_features.

    Returns
    -------
    DataFrame

    """
    roi_index_list = list()
    peak_index_list = list()
    mz_mean_list = list()
    mz_std_list = list()
    descriptors_list = list()
    roi = _load_roi(file_path)

    for roi_index, k_roi in enumerate(roi):
        k_roi.fill_nan()
        try:
            k_params = k_roi.find_peaks(smoothing_strength=smoothing_strength,
                                        descriptors=descriptors,
                                        filters=filters,
                                        find_peaks_params=find_peaks_params)
        except:
            print(str(file_path))
            print(roi_index)
            raise ValueError

        n_features = len(k_params)
        descriptors_list.extend(k_params)
        k_mz_mean, k_mz_std = k_roi.get_peaks_mz()
        roi_index_list.append([roi_index] * n_features)
        peak_index_list.append(range(n_features))
        mz_mean_list.append(k_mz_mean)
        mz_std_list.append(k_mz_std)
    _save_roi(file_path, roi)

    roi_index_list = np.hstack(roi_index_list)
    peak_index_list = np.hstack(peak_index_list)
    mz_mean_list = np.hstack(mz_mean_list)
    mz_std_list = np.hstack(mz_std_list)

    ft_table = pd.DataFrame(data=descriptors_list)
    ft_table = ft_table.rename(columns={"loc": "rt"})
    ft_table["mz"] = mz_mean_list
    ft_table["mz std"] = mz_std_list
    ft_table["roi index"] = roi_index_list
    ft_table["peak index"] = peak_index_list
    ft_table = ft_table.dropna(axis=0)
    ft_table["roi index"] = ft_table["roi index"].astype(int)
    ft_table["peak index"] = ft_table["peak index"].astype(int)
    # TODO: save features into a file
    # save dataframe into a file
    data_path = file_path.parent.parent
    df_path = data_path.joinpath("feature", file_path.stem + ".pickle")
    ft_table.to_pickle(df_path)
    return ft_table



"""
Functions and objects to work with mzML data and tabular data obtained from
third party software used to process Mass Spectrometry data.

Objects
-------
MSData: reads raw MS data in the mzML format. Creates Chromatograms, ROI and
MSSpectrum from ra data.

Functions
---------
read_pickle(path): Reads a DataContainer stored as a pickle.
read_progenesis(path): Reads data matrix in a csv file generated with
Progenesis software.
read_data_matrix(path, mode): Reads data matrix in several formats. Calls other
read functions.
functions.

See Also
--------
Chromatogram
MSSpectrum
DataContainer
Roi
"""

import os
import pickle
import requests
from typing import Optional, Callable, Union, List, BinaryIO, TextIO, Tuple
from typing import Generator
import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
from .container import DataContainer
from ._names import *
from . import lcms
from . import validation as v
from .utils import get_tidyms_path, gaussian_mixture, find_closest
from ._mzml import build_offset_list, get_spectrum, get_chromatogram


def read_pickle(path: Union[str, BinaryIO]) -> DataContainer:
    """
    read a DataContainer stored as a pickle

    Parameters
    ----------
    path: str or file
        path to read DataContainer

    Returns
    -------
    DataContainer
    """
    if hasattr(path, "read"):
        with path as fin:
            result = pickle.load(fin)
    else:
        with open(path, "rb") as fin:
            result = pickle.load(fin)
    return result


def read_progenesis(path: Union[str, TextIO]):
    """
    Read a progenesis file into a DataContainer

    Parameters
    ----------
    path : str or file

    Returns
    -------
    dc : DataContainer
    """
    df_header = pd.read_csv(path, low_memory=False)
    df = df_header.iloc[2:].copy()
    col_names = df_header.iloc[1].values
    df.columns = col_names
    df = df.set_index("Compound")
    df_header = df_header.iloc[:1].copy()
    df_header = df_header.fillna(axis=1, method="ffill")
    norm_index = df_header.columns.get_loc("Normalised abundance") - 1
    raw_index = df_header.columns.get_loc("Raw abundance") - 1
    ft_def = df.iloc[:, 0:norm_index].copy()
    data = df.iloc[:, raw_index:(2 * raw_index - norm_index)].T
    sample_info = \
        df_header.iloc[:, (raw_index + 1):(2 * raw_index - norm_index + 1)].T

    # rename data matrix
    data.index.rename("sample", inplace=True)
    data.columns.rename("feature", inplace=True)
    data = data.astype(float)

    # rename sample info
    sample_info.index = data.index
    sample_info.rename({sample_info.columns[0]: _sample_class},
                       axis="columns", inplace=True)

    # rename features def
    ft_def.index.rename("feature", inplace=True)
    ft_def.rename({"m/z": "mz", "Retention time (min)": "rt"},
                  axis="columns",
                  inplace=True)
    ft_def = ft_def.astype({"rt": float, "mz": float})
    ft_def["rt"] = ft_def["rt"] * 60
    v.validate_data_container(data, ft_def, sample_info)
    dc = DataContainer(data, ft_def, sample_info)
    return dc


def read_mzmine(data: Union[str, TextIO],
                sample_metadata: Union[str, TextIO]) -> DataContainer:
    """
    read a MZMine2 csv file into a DataContainer.

    Parameters
    ----------
    data : str or file
        csv file generated with MZMine.
    sample_metadata : str, file or DataFrame
        csv file with sample metadata. The following columns are required:
        * sample : the same sample names used in `data`
        * class : the sample classes
        Columns with run order and analytical batch information are optional.

    Returns
    -------
    DataContainer

    """
    df = pd.read_csv(data)
    col_names = pd.Series(df.columns)
    sample_mask = ~df.columns.str.startswith("row")

    # remove filename extensions
    col_names = col_names.str.split(".").apply(lambda x: x[0])
    # remove "row " in columns
    col_names = col_names.str.replace("row ", "")
    col_names = col_names.str.replace("Peak area", "")
    col_names = col_names.str.strip()
    df.columns = col_names
    df = df.rename(columns={"m/z": "mz", "retention time": "rt"})
    ft_metadata = df.loc[:, ["mz", "rt"]]

    # make feature metadata
    n_ft = df.shape[0]
    ft_len = len(str(n_ft))
    ft_names = ["FT" + str(x).rjust(ft_len, "0") for x in (range(1, n_ft + 1))]
    ft_metadata.index = ft_names

    # data matrix
    data_matrix = df.loc[:, sample_mask]
    data_matrix = data_matrix.loc[:, ~data_matrix.isna().all()]
    data_matrix.index = ft_names
    data_matrix = data_matrix.T

    if not isinstance(sample_metadata, pd.DataFrame):
        sample_metadata = pd.read_csv(sample_metadata)
        sample_metadata = sample_metadata.set_index("sample")

    dc = DataContainer(data_matrix, ft_metadata, sample_metadata)
    return dc


def read_xcms(data_matrix: str, feature_metadata: str,
              sample_metadata: str, class_column: str = "class",
              sep: str = "\t"):
    """
    Reads tabular data generated with xcms.

    Parameters
    ----------
    data_matrix : str
        Path to a tab-delimited data matrix generated with the R package
        SummarizedExperiment assay method.
    feature_metadata : str
        Path to a tab-delimited  file with feature metadata, (called feature
        definitions in XCMS) generated with the R package SummarizedExperiment
        colData method.
    sample_metadata : str
        Path to a tab-delimited  file with sample metadata, generated with the
        R package SummarizedExperiment colData method. A column named class is
        required. If the class information is under another name, it must be
        specified in the `class_column` parameter.
    class_column : str
        Column name which holds sample class information in the sample metadata.
    sep : str
        Separator used in the files. As the feature metadata generated with XCMS
        has comma characters, the default value for the separator is "\t".

    Returns
    -------
    DataContainer
    """

    # data matrix
    dm = pd.read_csv(data_matrix, sep=sep)
    dm = dm.T
    dm.columns.name = "feature"
    dm.index.name = "sample"

    # feature metadata
    fm = pd.read_csv(feature_metadata, sep=sep)
    fm.index.name = "feature"
    fm = fm.rename(columns={"mzmed": "mz", "rtmed": "rt"})
    fm = fm.loc[:, ["mz", "rt"]]
    # TODO : include information from CAMERA package

    # sample_metadata
    sm = pd.read_csv(sample_metadata, sep=sep, index_col=0)
    sm.index.name = "sample"
    sm = sm.rename(columns={class_column: "class"})
    dc = DataContainer(dm, fm, sm)
    return dc


def read_data_matrix(path: Union[str, TextIO, BinaryIO],
                     data_matrix_format: str,
                     sample_metadata: Optional[str] = None
                     ) -> DataContainer:
    """
    Read different Data Matrix formats into a DataContainer.

    Parameters
    ----------
    path: str
        path to the data matrix file.
    data_matrix_format: {"progenesis", "pickle", "mzmine"}
    sample_metadata : str, file or DataFrame.
        Required for mzmine data.

    Returns
    -------
    DataContainer

    Examples
    --------
    >>> data = read_data_matrix("data_path.csv", "progenesis")
    """
    if data_matrix_format == "progenesis":
        return read_progenesis(path)
    elif data_matrix_format == "pickle":
        return read_pickle(path)
    elif data_matrix_format == "mzmine":
        return read_mzmine(path, sample_metadata)
    else:
        msg = "Invalid Format"
        raise ValueError(msg)


class MSData:
    """
    Reader object for raw MS data.

    Manages chromatogram, roi and spectra creation.

    Attributes
    ----------
    path : str
        Path to a mzML file.
    ms_mode : {"centroid", "profile"}, default="centroid"
        The mode in which the MS data is stored.
    instrument : {"qtof". "orbitrap"}, default="qtof"
        The MS instrument type used to acquire the experimental data. Used to
        set default parameters in the methods.
    separation : {"uplc", "hplc"}, default="uplc"
        The separation technique used before MS analysis. Used to
        set default parameters in the methods.

    """

    def __init__(self, path: str, ms_mode: str = "centroid",
                 instrument: str = "qtof", separation: str = "uplc"):
        sp_offset, chrom_offset, index_offset = build_offset_list(path)
        self._spectra_offset = sp_offset
        self._chromatogram_offset = chrom_offset
        self._index_offset = index_offset
        self.path = path

        self.ms_mode = ms_mode
        self.instrument = instrument
        self.separation = separation

    @property
    def ms_mode(self) -> str:
        return self._ms_mode

    @ms_mode.setter
    def ms_mode(self, value: Optional[str]):
        if value in ["centroid", "profile"]:
            self._ms_mode = value
        else:
            msg = "mode must be `centroid` or `profile`. Got `{}`."
            raise ValueError(msg.format(value))

    @property
    def instrument(self) -> str:
        return self._instrument

    @instrument.setter
    def instrument(self, value: str):
        valid_instruments = ["qtof", "orbitrap"]
        if value in valid_instruments:
            self._instrument = value
        else:
            msg = "`instrument` must be `orbitrap` or `qtof`. Got `{}`."
            msg = msg.format(value)
            raise ValueError(msg)

    @property
    def separation(self) -> str:
        return self._separation

    @separation.setter
    def separation(self, value: str):
        valid_separation = ["uplc", "hplc"]
        if value in valid_separation:
            self._separation = value
        else:
            msg = "`separation` must be any of {}".format(valid_separation)
            raise ValueError(msg)

    def get_n_chromatograms(self) -> int:
        """
        Get the chromatogram count in the file

        Returns
        -------
        n_chromatograms : int

        """
        return len(self._chromatogram_offset)

    def get_n_spectra(self) -> int:
        """
        Get the spectra count in the file

        Returns
        -------
        n_spectra : int

        """
        return len(self._spectra_offset)

    def get_chromatogram(self, n: int) -> Tuple[str, lcms.Chromatogram]:
        """
        Get the nth chromatogram stored in the file.

        Parameters
        ----------
        n : int

        Returns
        -------
        name : str
        chromatogram : lcms.Chromatogram

        """
        chrom_data = get_chromatogram(
            self.path,
            self._spectra_offset,
            self._chromatogram_offset,
            self._index_offset,
            n
        )
        name = chrom_data["name"]
        chromatogram = lcms.Chromatogram(
            chrom_data["time"], chrom_data["spint"], mode=self.separation)
        return name, chromatogram

    def get_spectrum(self, scan: int) -> lcms.MSSpectrum:
        """
        get the spectrum for the given scan number.

        Parameters
        ----------
        scan: int
            scan number

        Returns
        -------
        MSSpectrum

        """
        sp_data = get_spectrum(
            self.path,
            self._spectra_offset,
            self._chromatogram_offset,
            self._index_offset,
            scan
        )
        sp_data["is_centroid"] = self.ms_mode == "centroid"
        return lcms.MSSpectrum(**sp_data)

    @v.validated_ms_data(v.spectra_iterator_schema)
    def get_spectra_iterator(
            self,
            ms_level: int = 1,
            start: int = 0,
            end: Optional[int] = None,
            start_time: float = 0.0,
            end_time: Optional[float] = None
    ) -> Generator[Tuple[int, lcms.MSSpectrum], None, None]:
        """
        Yields the spectra in the file.

        Parameters
        ----------
        ms_level : int, default=1
            Use data from this ms level.
        start : int, default=0
            Starts iteration at this scan number.
        end : int or None, default=None
            Ends iteration at this scan number. By default, uses all scans.
        start_time : float, default=0.0
            Ignore scans with acquisition times lower than this value.
        end_time : float or None, default=None
            Ignore scans with acquisition times higher than this value.

        Yields
        ------
        scan_number: int
        spectrum : lcms.MSSpectrum

        """
        for k in range(start, end):
            sp = self.get_spectrum(k)
            is_valid_level = ms_level == sp.ms_level
            is_valid_time = ((start_time <= sp.time) and
                             ((end_time is None) or (end_time > sp.time)))
            if is_valid_level and is_valid_time:
                yield k, sp

    @v.validated_ms_data(v.make_chromatogram_schema)
    def make_chromatograms(
            self,
            mz: np.ndarray,
            window: Optional[float] = None,
            accumulator: str = "sum",
            ms_level: int = 1,
            start: int = 0,
            end: Optional[int] = None,
            start_time: float = 0.0,
            end_time: Optional[float] = None,
    ) -> List[lcms.Chromatogram]:
        """
        Computes extracted ion chromatograms using a list of m/z values.

        Parameters
        ----------
        mz : array
            m/z values used to build the EICs.
        window : positive number or None, default=None
            m/z tolerance used to build the EICs. If ``self.instrument`` is
            ``"qtof"``, the default value is ``0.05``. If ``self.instrument`` is
            ``"orbitrap"`` the default value is ``0.005``.
        accumulator : {"sum", "mean"}, default="sum"
            Mode used to accumulate the values inside the m/z window. ``"sum"``
            computes the total intensity inside the window. ``"mean"`` divides
            the total intensity using the number of points inside the window.
        ms_level : int, default=1
            ms level used to build the chromatograms.
        start : int, default=0
            Create chromatograms starting at this scan number.
        end : int or None, default=None
            End chromatograms at this scan number. By default, uses all scans.
        start_time : float, default=0.0
            include scans starting at this acquisition time.
        end_time : float or None, default=None
            Stops when the acquisition time is higher than this value.

        Returns
        -------
        chromatograms : List of Chromatograms

        """
        n_sp = self.get_n_spectra()

        # mz_intervals has this shape to be compatible with reduce at
        mz_intervals = (np.vstack((mz - window, mz + window))
                        .T.reshape(mz.size * 2))

        eic = np.zeros((mz.size, n_sp))
        rt = np.zeros(n_sp)
        valid_index = list()
        sp_iterator = self.get_spectra_iterator(
            ms_level, start, end, start_time, end_time)
        for scan, sp in sp_iterator:
            valid_index.append(scan)
            rt[scan] = sp.time
            sp_size = sp.mz.size

            # prevents error when working with empty spectra
            if sp_size == 0:
                continue

            # values for each eic in the current scan
            ind_sp = np.searchsorted(sp.mz, mz_intervals)  # slices for each eic
            has_mz = (ind_sp[1::2] - ind_sp[::2]) > 0  # find non-empty slices
            # elements added at the end of mz_sp raise IndexError
            ind_sp[ind_sp >= sp_size] = sp_size - 1
            # this adds the values between two consecutive indices
            tmp_eic = np.where(
                has_mz, np.add.reduceat(sp.spint, ind_sp)[::2], 0)
            if accumulator == "mean":
                norm = ind_sp[1::2] - ind_sp[::2]
                norm[norm == 0] = 1
                tmp_eic = tmp_eic / norm
            eic[:, scan] = tmp_eic
        valid_index = np.array(valid_index)
        rt = rt[valid_index]
        eic = eic[:, valid_index]

        chromatograms = list()
        for row in eic:
            chromatogram = lcms.Chromatogram(
                rt.copy(), row, mode=self.separation)
            chromatograms.append(chromatogram)
        return chromatograms

    def make_tic(self,
                 kind: str = "tic",
                 ms_level: int = 1,
                 start: int = 0,
                 end: Optional[int] = None,
                 start_time: float = 0.0,
                 end_time: Optional[float] = None,
                 ) -> lcms.Chromatogram:
        """
        Creates a total ion chromatogram.

        Parameters
        ----------
        kind: {"tic", "bpi"}, default="tic"
            `tic` computes the total ion chromatogram. `bpi` computes the base
            peak chromatogram.
        ms_level : int, default=1
            ms level used to build the chromatogram.
        start : int, default=0
            Create the chromatogram starting at this scan
        end : int or None, default=None
            End chromatograms at this scan. By default, use all scans.
        start_time : float, default=0.0
            include scans starting at this acquisition time.
        end_time : float or None, default=None
            Stops when the acquisition time is higher than this value. If None,
            it doesn't filter scans by time.

        Returns
        -------
        chromatograms : lcms.Chromatograms

        """
        if kind == "tic":
            reduce = np.sum
        elif kind == "bpi":
            reduce = np.max
        else:
            msg = "valid modes are tic or bpi"
            raise ValueError(msg)

        n_scan = self.get_n_spectra()
        rt = np.zeros(n_scan)
        tic = np.zeros(n_scan)
        valid_index = list()
        # it is not possible to know a priori how many scans of each level are
        # available in a given file without iterating over it. valid_index holds
        # the index related to the selected level and is used to remove scans
        # from other levels.
        sp_iterator = self.get_spectra_iterator(
            ms_level, start, end, start_time, end_time)
        for scan, sp in sp_iterator:
            valid_index.append(scan)
            rt[scan] = sp.time
            if sp.spint.size:
                tic[scan] = reduce(sp.spint)
            else:
                tic[scan] = 0
        tic = tic[valid_index]
        rt = rt[valid_index]
        return lcms.Chromatogram(rt, tic, self.separation)

    @v.validated_ms_data(v.accumulate_spectra_schema)
    def accumulate_spectra(
        self,
        start: int,
        end: int,
        subtract_left: Optional[int] = None,
        subtract_right: Optional[int] = None,
        ms_level: int = 1
    ) -> lcms.MSSpectrum:
        """
        accumulates a series of consecutive spectra into a single spectrum.

        Parameters
        ----------
        start: int
            Start accumulating scans at this scan number.
        end: int
            Ends accumulation at this scan number. Scans in the range
            [start:end) are used
        subtract_left : int or None, default=None
            Scans in the range [subtract_left:start) are subtracted from the
            accumulated spectrum. If ``None``, don't subtract anything.
        subtract_right : int, or None, default=None
            Scans in the range [end:subtract_right) are subtracted from the
            accumulated spectrum. If ``None``, don't subtract anything.
        ms_level : int, default=1
            ms level used to build the accumulated spectrum.

        Returns
        -------
        MSSpectrum

        """
        if self.ms_mode == "centroid":
            sp = self._accumulate_spectra_centroid(
                start, end, subtract_left, subtract_right, ms_level)
        else:   # profile
            sp = self._accumulate_spectra_profile(
                start, end, subtract_left, subtract_right, ms_level)
        return sp

    def get_rt(
            self,
            start: int = 0,
            end: Optional[int] = None,
            ms_level: int = 1,
            start_time: float = 0.0,
            end_time: Optional[float] = None
    ) -> np.ndarray:
        """
        Gets the retention time array for the experiment.

        Parameters
        ----------
        start : int, default=0
            include values starting from this scan number.
        end : int or None, default=None
            include values until this scan number. If None, ends after the last
            scan.
        ms_level : int, default=1
            Use scans from this ms level.
        start_time : float, default=0.0
            Use scans with acquisition time greater than this value.
        end_time : float or None, defaults=None
            If specified, includes scans with acquisition times lower than
            this value.

        Returns
        -------
        rt : np.ndarray

        ms_level : int, optional
            data level to use. If None returns a retention time array for all
            levels.

        """
        sp_iter = self.get_spectra_iterator(
            ms_level, start, end, start_time, end_time)
        rt = list()
        for _, sp in sp_iter:
            rt.append(sp.time)
        return np.array(rt)

    @v.validated_ms_data(v.make_roi_schema)
    def make_roi(
        self,
        tolerance: Optional[float] = None,
        max_missing: Optional[int] = None,
        min_length: Optional[int] = None,
        min_intensity: float = 0.0,
        multiple_match: str = "reduce",
        mz_reduce: Union[str, Callable] = "mean",
        sp_reduce: Union[str, Callable] = "sum",
        targeted_mz: Optional[np.ndarray] = None,
        pad: Optional[int] = None,
        ms_level: int = 1,
        start: int = 0,
        end: Optional[int] = None,
        start_time: float = 0.0,
        end_time: Optional[float] = None,
        min_snr: float = 10,
        min_distance: Optional[float] = None
    ) -> List[lcms.Roi]:
        """
        Builds regions of interest from raw data.

        See the Notes for a detailed description of the algorithm used.

        Parameters
        ----------
        tolerance : positive number or None, default=None
            m/z tolerance to connect values across scans. If None, the value is
            set based on the value of ``self.instrument``. If
            ``self.instrument`` is ``"qtof"``, the tolerance is ``0.01``. If
            ``self.instrument`` is ``"orbitrap"``, the tolerance is ``0.005``
        max_missing : non-negative integer or None, default=None
            maximum number of consecutive missing values in a valid  ROI. If
            ``None``, the value is set to ``1``.
        min_length : positive integer or None, default=None
            The minimum length of a valid ROI. If, ``None``, the value is set
            based on ``self.separation``. If ``self.separation`` is ``"uplc"``,
            the value is set to ``10``. If ``self.separation`` is ``"hplc"``,
            the value is set to ``20``.
        min_intensity : non-negative number , default=0.0
            Minimum intensity in a valid ROI.
        pad: int or None, default=None
            Pad dummy values to the left and right of the ROI. This produces
            better peak picking results when searching low intensity peaks in a
            ROI. Using None set the value to ``2`` if ``self.separation`` is
            ``"uplc"`` or ``"hplc"``.
        multiple_match : {"closest", "reduce"}, default="reduce"
            How peaks are matched when there is more than one valid match. If
            ``"closest"`` is used, the closest peak is assigned as a match and
            the others are used to create new ROIs. If ``"reduce"`` is used,
            unique m/z and intensity values are generated using the reduce
            function in `mz_reduce` and `sp_reduce` respectively.
        mz_reduce : "mean" or Callable, default="mean"
            Function used to reduce m/z values. If ``"mean"`` is used, the mean
            value of all valid m/z is used. Any function that accepts numpy
            arrays and return numbers can be used. Used only when
            `multiple_match` is set to ``"reduce"``. See the following
            prototype:

            .. code-block:: python

                def mz_reduce(mz_match: np.ndarray) -> float:
                    pass

        sp_reduce : {"mean", "sum"} or Callable, default="sum"
            Function used to reduce intensity values. ``"mean"`` computes the
            mean intensity and ``"sum"`` computes the total intensity. Any
            function that accepts numpy arrays and return numbers can be used.
            Only used when `multiple_match` is set to ``"reduce"``. See the
            prototype shown on `mz_reduce`.
        targeted_mz : numpy.ndarray or None, default=None
            A list of m/z values to perform a targeted ROI creation. If this
            value is provided, only ROI with these m/z values will be created.
        ms_level : int, default=1
            ms level used to build the ROI.
        start : int, default=0
            Create ROI starting at this scan
        end : int or None, default=None
            Stop ROI creation at this scan. If None, stops after the last scan
            in the file.
        start_time : float, default=0.0
            Use scans starting at this acquisition time.
        end_time : float or None, default=None
            Stops when the acquisition time is higher than this value.
        min_snr : positive number, default=10.0
            Minimum signal-to-noise ratio of the peaks. Used only to convert
            profile data to centroid mode
        min_distance : positive number or None, default=None
            Minimum distance between consecutive peaks. If ``None``, the value
            is set to 0.01 if ``self.instrument`` is ``"qtof"`` or to 0.005 if
            ``self.instrument`` is ``"orbitrap"``. Used only to convert profile
            data to centroid mode.

        Notes
        -----
        ROIs are built using the method described in [TR08]_ with slight
        modifications.

        A ROI is modelled as a combination of three arrays with the same size:
        m/z, intensity and time. ROIs are created and extended connecting
        m/z values across successive scans using the following method:

        1.  The m/z values in The first scan are used to initialize a list of
            ROI. If `targeted_mz` is used, the ROI are initialized using this
            list.
        2.  m/z values from the next scan extend the ROIs if they are closer
            than `tolerance` to the mean m/z of the ROI. Values that don't match
            any ROI are used to create new ROIs and are appended to the ROI
            list (only if `targeted_mz` is not used).
        3.  If more than one m/z value is within the tolerance threshold,
            m/z and intensity values are computed according to the
            `multiple_match` strategy.
        4.  If a ROI can't be extended with any m/z value from the new scan,
            it is extended using NaNs.
        5.  If the last `max_missing` values of a ROI are NaN, the ROI is
            flagged as finished. If the maximum intensity of a finished ROI
            is greater than `min_intensity` and the number of points is greater
            than `min_length`, then the ROI is flagged as valid. Otherwise, the
            ROI is discarded.
        6.  Repeat from step 2 until no more scans are available.

        Returns
        -------
        roi : list[Roi]
            A list with the detected regions of interest.

        Raises
        ------
        ValueError
            If the data is not in centroid mode.

        See Also
        --------
        lcms.Roi : Representation of a ROI.

        References
        ----------
        .. [TR08] Tautenhahn, R., Böttcher, C. & Neumann, S. Highly sensitive
            feature detection for high resolution LC/MS. BMC Bioinformatics 9,
            504 (2008). https://doi.org/10.1186/1471-2105-9-504

        """

        if targeted_mz is None:
            sp_seed = self.get_spectrum(start)
            mz_seed, _ = sp_seed.find_centroids(min_snr, min_distance)
            targeted = False
        else:
            mz_seed = targeted_mz
            targeted = True

        rt = np.zeros(self.get_n_spectra())
        processor = lcms.RoiMaker(
            mz_seed,
            max_missing=max_missing,
            min_length=min_length,
            min_intensity=min_intensity,
            tolerance=tolerance,
            multiple_match=multiple_match,
            mz_reduce=mz_reduce,
            sp_reduce=sp_reduce,
            targeted=targeted
        )

        scans = list()  # scan number used in to build ROI
        sp_iterator = self.get_spectra_iterator(
            ms_level,
            start,
            end,
            start_time,
            end_time
        )
        for scan, sp in sp_iterator:
            rt[scan] = sp.time
            scans.append(scan)
            sp_mz, sp_spint = sp.find_centroids(min_snr, min_distance)
            processor.extend_roi(sp_mz, sp_spint, scan)
            processor.store_completed_roi()

        # add roi not completed during the last scan
        processor.flag_as_completed()
        processor.store_completed_roi()

        # extend roi, find rt of each roi and convert to Roi objects
        roi = processor.process_completed_roi(scans, rt, pad, self.separation)
        return roi

    def _accumulate_spectra_centroid(
        self,
        start: int,
        end: int,
        subtract_left: int,
        subtract_right: int,
        ms_level: int = 1
    ) -> lcms.MSSpectrum:
        """
        accumulates a series of consecutive spectra into a single spectrum.

        auxiliary method for accumulate_spectra.

        """
        # don't remove any m/z value when detecting rois
        max_missing = subtract_right - subtract_left

        roi = self.make_roi(
            max_missing=max_missing,
            min_length=1,
            start=subtract_left,
            end=subtract_right,
            ms_level=ms_level
        )

        mz = np.zeros(len(roi))
        spint = mz.copy()

        # set subtract values to negative
        for k, r in enumerate(roi):
            accum_mask = - np.ones(r.scan.size)
            accum_start, accum_end = np.searchsorted(r.scan, [start, end])
            accum_mask[accum_start:accum_end] = 1
            mz[k] = np.nanmean(r.mz)
            spint[k] = np.nansum(r.spint * accum_mask)

        # remove negative values
        pos_values = spint > 0
        mz = mz[pos_values]
        spint = spint[pos_values]

        # sort values
        sorted_index = np.argsort(mz)
        mz = mz[sorted_index]
        spint = spint[sorted_index]

        sp = lcms.MSSpectrum(
            mz, spint, ms_level=ms_level, instrument=self.instrument)
        return sp

    def _accumulate_spectra_profile(
        self,
        start: int,
        end: int,
        subtract_left: int,
        subtract_right: int,
        ms_level: int = 1,
    ) -> lcms.MSSpectrum:
        """
        aux method for accumulate_spectra.

        """
        # The spectra are accumulated in two steps:
        #
        #  1.  iterate through scans to build a grid of m/z values for the
        #      accumulated spectra.
        #  2.  A second iteration is done to interpolate the intensity in each
        #      scan to the m/z grid and generate the accumulated spectrum.
        #
        #  This process is done in two steps to avoid storing the intensity
        #  values from each scan until the grid is built.

        accum_mz = None
        # m/z tol. A small value is used to prevent distortions in the results
        tol = 0.00005
        sp_iter = self.get_spectra_iterator(
            ms_level,
            subtract_left,
            subtract_right
        )
        # first iteration. Builds a grid of m/z values for the accumulated
        # spectrum. The grid is extended using new m/z values that appear
        # in each new scan
        for scan, sp in sp_iter:
            if accum_mz is None:
                accum_mz = sp.mz
            ind = find_closest(accum_mz, sp.mz)
            no_match = np.abs(accum_mz[ind] - sp.mz) > tol
            accum_mz = np.sort(np.hstack((accum_mz, sp.mz[no_match])))

        accum_sp = np.zeros_like(accum_mz)
        sp_iter = self.get_spectra_iterator(
            ms_level,
            subtract_left,
            subtract_right
        )

        for scan, sp in sp_iter:
            interpolator = interp1d(sp.mz, sp.spint, fill_value=0.0)
            if (scan < start) or (scan > end):
                sign = -1
            else:
                sign = 1
            accum_sp += interpolator(accum_mz) * sign

        # set negative values that may result from subtraction to zero
        is_positive_sp = accum_sp > 0
        accum_mz = accum_mz[is_positive_sp]
        accum_sp = accum_sp[is_positive_sp]

        res = lcms.MSSpectrum(
            accum_mz,
            accum_sp,
            instrument=self.instrument,
            ms_level=ms_level,
            is_centroid=False
        )
        return res


class SimulatedMSData(MSData):  # pragma: no cover
    """
    Emulates a MSData using simulated data. Used for tests.

    """
    def __init__(self, mz_values: np.ndarray, rt_values: np.ndarray,
                 mz_params: np.ndarray, rt_params: np.ndarray,
                 ft_noise: Optional[np.ndarray] = None,
                 noise: Optional[float] = None, ms_mode: str = "centroid",
                 separation: str = "uplc", instrument: str = "qtof"):
        """
        Constructor function

        Parameters
        ----------
        mz_values : array
            sorted mz values for each mz scan, used to build spectra in profile
            mode.
        rt_values : array
            sorted rt values, used to set scan time.
        mz_params : array with shape (n, 3)
             Used to build m/z peaks. Each row is the mean, standard deviation
             and amplitude in the m/z dimension.
        rt_params : array with shape (n, 3)
             Used to build rt peaks. Each row is the mean, standard deviation
             and amplitude in the rt dimension
        ft_noise : array_with shape (n, 2), optional
            adds noise to mz and rt values
        noise : positive number, optional
            noise level to add to each scan. the noise is modeled as gaussian
            iid noise in each scan with standard deviation equal to the value
            used. An offset value is added to make the noise contribution
            non-negative. To make the noise values reproducible, each scan has
            a seed value associated to generate always the same noise value in
            a given scan. seed values are generated randomly each time a new
            object is instantiated.

        """
        # MSData params
        self._spectra_offset = None
        self._chromatogram_offset = None
        self._index_offset = None
        self.path = None

        self.ms_mode = ms_mode
        self.instrument = instrument
        self.separation = separation

        # simulation params
        self.mz = mz_values
        self.mz_params = mz_params
        self.rt = rt_values
        self.n_scans = rt_values.size
        self._seeds = None
        self._noise_level = None
        self.ft_noise = ft_noise
        # seeds are used to ensure that each time that a spectra is generated
        # with get_spectra its values are the same
        self._seeds = np.random.choice(self.n_scans * 10, self.n_scans)
        self._noise_level = noise

        if ft_noise is not None:
            np.random.seed(self._seeds[0])
            rt_params[0] += np.random.normal(scale=ft_noise[1])
        self.rt_array = gaussian_mixture(rt_values, rt_params)

    def get_n_spectra(self):
        return self.n_scans

    def get_spectrum(self, scan_number: int):
        is_valid_scan = (0 <= scan_number) and (self.n_scans > scan_number)
        if not is_valid_scan:
            msg = "Invalid scan number."
            raise ValueError(msg)
        rt = self.rt[scan_number]

        # create a mz array for the current scan
        if self.ft_noise is not None:
            np.random.seed(self._seeds[scan_number])
            mz_noise = np.random.normal(scale=self.ft_noise[0])
            mz_params = self.mz_params.copy()
            mz_params[:, 0] += mz_noise
        else:
            mz_params = self.mz_params

        if self.ms_mode == "centroid":
            mz = mz_params[:, 0]
            spint = self.rt_array[:, scan_number] * mz_params[:, 1]
        else:
            mz = self.mz
            mz_array = gaussian_mixture(self.mz, mz_params)
            spint = self.rt_array[:, scan_number][:, np.newaxis] * mz_array
            spint = spint.sum(axis=0)

        if self._noise_level is not None:
            np.random.seed(self._seeds[scan_number])
            noise = np.random.normal(size=mz.size, scale=self._noise_level)
            noise -= noise.min()
            spint += noise
        sp = lcms.MSSpectrum(
            mz, spint, time=rt, ms_level=1, instrument=self.instrument,
            is_centroid=(self.ms_mode == "centroid"))
        return sp


def list_available_datasets(hide_test_data: bool = True) -> List[str]:
    """
    List available example datasets

    Parameters
    ----------
    hide_test_data : bool

    Returns
    -------
    datasets: List[str]
    """
    repo_url = "https://raw.githubusercontent.com/griquelme/tidyms-data/master/"
    dataset_url = repo_url + "datasets.txt"
    r = requests.get(dataset_url)
    datasets = r.text.split()
    if hide_test_data:
        datasets = [x for x in datasets if not x.startswith("test")]
    return datasets


def _check_dataset_name(name: str):
    available = list_available_datasets(False)
    if name not in available:
        msg = "Invalid dataset name. Available datasets are: {}"
        msg = msg.format(available)
        raise ValueError(msg)


def _download_dataset(name: str):
    """Download a dataset from GitHub"""
    _check_dataset_name(name)
    cache_path = get_tidyms_path()
    dataset_path = os.path.join(cache_path, name)
    if not os.path.exists(dataset_path):
        os.makedirs(dataset_path)
    repo_url = "https://raw.githubusercontent.com/griquelme/tidyms-data/master/"
    dataset_url = repo_url + "/" + name
    files = ["feature.csv", "sample.csv", "data.csv"]
    for f in files:
        file_url = dataset_url + "/" + f
        r = requests.get(file_url)
        file_path = os.path.join(dataset_path, f)
        with open(file_path, "w") as fin:
            fin.write(r.text)


def load_dataset(name: str, cache: bool = True, **kwargs) -> DataContainer:
    """
    load example dataset into a DataContainer. Available datasets can be seen
    using the list_datasets function.

    Parameters
    ----------
    name : str
        name of an available dataset.
    cache : bool
        If True tries to read the dataset from a local cache.
    kwargs: additional parameters to pass to the Pandas read_csv function

    Returns
    -------
    data_matrix : DataFrame
    feature_metadata : DataFrame
    sample_metadata : DataFrame

    """
    cache_path = get_tidyms_path()
    dataset_path = os.path.join(cache_path, name)
    sample_path = os.path.join(dataset_path, "sample.csv")
    feature_path = os.path.join(dataset_path, "feature.csv")
    data_matrix_path = os.path.join(dataset_path, "data.csv")

    is_data_found = (os.path.exists(sample_path) and
                     os.path.exists(feature_path) and
                     os.path.exists(data_matrix_path))

    if not (cache and is_data_found):
        _download_dataset(name)
        pass

    sample_metadata = pd.read_csv(sample_path, index_col=0, **kwargs)
    feature_metadata = pd.read_csv(feature_path, index_col=0, **kwargs)
    data_matrix = pd.read_csv(data_matrix_path, index_col=0, **kwargs)
    dataset = DataContainer(data_matrix, feature_metadata, sample_metadata)
    return dataset



"""
Functions and objects for working with LC-MS data read from pyopenms.

Objects
-------
Chromatogram
MSSpectrum
Roi

Functions
---------
make_chromatograms
make_roi
accumulate_spectra_profile
accumulate_spectra_centroid
get_lc_filter_peak_params
get_roi_params
get_find_centroid_params

"""

import bokeh.plotting
import numpy as np
from collections import namedtuple, deque
from scipy.ndimage import gaussian_filter1d
from typing import Optional, Tuple, Union, List, Callable
from scipy.interpolate import interp1d
from . import peaks
from . import _plot_bokeh
from .utils import find_closest


class MSSpectrum:
    """
    Representation of a Mass Spectrum. Manages conversion to
    centroid and plotting of data.

    Attributes
    ----------
    mz : array
        m/z data
    spint : array
        Intensity data
    time : float or None
        Time at which the spectrum was acquired
    ms_level : int
        MS level of the scan
    polarity : int or None
        Polarity used to acquire the data.
    instrument : {"qtof", "orbitrap"}, default="qtof"
        MS instrument type. Used to set default values in methods.
    is_centroid : bool
        True if the data is in centroid mode.

    """

    def __init__(
        self,
        mz: np.ndarray,
        spint: np.ndarray,
        time: Optional[float] = None,
        ms_level: int = 1,
        polarity: Optional[int] = None,
        instrument: str = "qtof",
        is_centroid: bool = True
    ):
        self.mz = mz
        self.spint = spint
        self.time = time
        self.ms_level = ms_level
        self.polarity = polarity
        self.instrument = instrument
        self.is_centroid = is_centroid

    @property
    def instrument(self) -> str:
        return self._instrument

    @instrument.setter
    def instrument(self, value):
        valid_values = ["qtof", "orbitrap"]
        if value in valid_values:
            self._instrument = value
        else:
            msg = "instrument must be one of `qtof` or `orbitrap`. Got {}"
            raise ValueError(msg.format(value))

    def find_centroids(
        self,
        min_snr: float = 10.0,
        min_distance: Optional[float] = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        r"""
        Find centroids in the spectrum.

        Parameters
        ----------
        min_snr : positive number, default=10.0
            Minimum signal-to-noise ratio of the peaks.
        min_distance : positive number or None, default=None
            Minimum distance between consecutive peaks. If ``None``, the value
            is set to 0.01 if ``self.instrument`` is ``"qtof"`` or to 0.005 if
            ``self.instrument`` is ``"orbitrap"``.

        Raises
        ------
        ValueError: if the spectrum is in centroid mode

        Returns
        -------
        centroid : array
            m/z centroids. If ``self.is_centroid`` is ``True``, returns
            ``self.mz``.
        area : array
            peak area. If ``self.is_centroid`` is ``True``, returns
            ``self.spint``.

        """
        if self.is_centroid:
            centroid, area = self.mz, self.spint
        else:
            params = get_find_centroid_params(self.instrument)
            if min_distance is not None:
                params["min_distance"] = min_distance

            if min_snr is not None:
                params["min_snr"] = min_snr

            centroid, area = peaks.find_centroids(self.mz, self.spint, **params)
        return centroid, area

    def plot(
        self,
        fig_params: Optional[dict] = None,
        line_params: Optional[dict] = None,
        show: bool = True
    ) -> bokeh.plotting.Figure:     # pragma: no cover
        """
        Plot the spectrum using Bokeh.

        Parameters
        ----------
        fig_params : dict or None, default=None
            key-value parameters to pass to ``bokeh.plotting.figure``.
        line_params : dict, or None, default=None
            key-value parameters to pass to ``bokeh.plotting.Figure.line``.
        show : bool, default=True
            If True calls ``bokeh.plotting.show`` on the Figure.
        Returns
        -------
        bokeh.plotting.Figure

        """
        default_fig_params = _plot_bokeh.get_spectrum_figure_params()
        if fig_params:
            default_fig_params.update(fig_params)
            fig_params = default_fig_params
        else:
            fig_params = default_fig_params
        fig = bokeh.plotting.Figure(**fig_params)

        if self.is_centroid:
            plotter = _plot_bokeh.add_stems
        else:
            plotter = _plot_bokeh.add_line
        plotter(fig, self.mz, self.spint, line_params=line_params)
        _plot_bokeh.set_ms_spectrum_axis_params(fig)
        if show:
            bokeh.plotting.show(fig)
        return fig


class Chromatogram:
    """
    Representation of a chromatogram. Manages plotting and peak detection.

    Attributes
    ----------
    rt : array
        Retention time data.
    spint : array
        Intensity data.
    mode : {"uplc", "hplc"}, default="uplc"
        Analytical platform used for separation. Sets default values for peak
        detection.

    """

    def __init__(self, rt: np.ndarray, spint: np.ndarray, mode: str = "uplc"):
        self.mode = mode
        self.rt = rt
        self.spint = spint
        self.peaks = None

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, value):
        valid_values = ["uplc", "hplc"]
        if value in valid_values:
            self._mode = value
        else:
            msg = "mode must be one of {}. Got {}"
            raise ValueError(msg.format(valid_values, value))

    def find_peaks(
        self,
        smoothing_strength: Optional[float] = 1.0,
        descriptors: Optional[dict] = None,
        filters: Optional[dict] = None,
        find_peaks_params: Optional[dict] = None,
        return_signal_estimators: bool = False
    ) -> List[dict]:
        """
        Find peaks and compute peak descriptors.

        Stores the detected peaks in the `peaks` attribute and returns the peaks
        descriptors.

        Parameters
        ----------
        descriptors : dict or None, default=None
            descriptors to pass to :py:func:`tidyms.peaks.get_peak_descriptors`.
        find_peaks_params : dict or None, default=None
            parameters to pass to :py:func:`tidyms.peaks.detect_peaks`
        smoothing_strength: positive number or None, default=1.0
            Width of a gaussian window used to smooth the signal. If ``None``,
            no smoothing is applied.
        filters : dict or None, default=None
            filters to pass to :py:func:`tidyms.peaks.get_peak_descriptors`
        return_signal_estimators : bool, default=False
            If True, returns a dictionary with the noise, baseline and the
            smoothed signal

        Returns
        -------
        params : List[dict]
            List of peak descriptors
        estimators : dict
            a dictionary with the noise, baseline and smoothed signal used
            inside the function. Only if `return_signal_estimators` is ``True``.

        Notes
        -----
        Peak detection is done in five steps:

        1. Estimate the noise level.
        2. Apply a gaussian smoothing to the chromatogram.
        3. Estimate the baseline.
        4. Detect peaks in the chromatogram.
        5. Compute peak descriptors and filter peaks.

        See Also
        --------
        peaks.estimate_noise : noise estimation of 1D signals
        peaks.estimate_baseline : baseline estimation of 1D signals
        peaks.detect_peaks : peak detection of 1D signals.
        peaks.get_peak_descriptors: computes peak descriptors.
        lcms.get_lc_filter_peaks_params : default value for filters

        """
        if find_peaks_params is None:
            find_peaks_params = dict()

        if filters is None:
            filters = get_lc_filter_peak_params(self.mode)

        noise = peaks.estimate_noise(self.spint)

        if smoothing_strength is None:
            x = self.spint
        else:
            x = gaussian_filter1d(self.spint, smoothing_strength)

        baseline = peaks.estimate_baseline(x, noise)
        peak_list = peaks.detect_peaks(x, noise, baseline, **find_peaks_params)

        peak_list, peak_descriptors = peaks.get_peak_descriptors(
            self.rt,
            self.spint,
            noise,
            baseline,
            peak_list,
            descriptors=descriptors,
            filters=filters
        )
        self.peaks = peak_list

        if return_signal_estimators:
            estimators = {"smoothed": x, "noise": noise, "baseline": baseline}
            res = peak_descriptors, estimators
        else:
            res = peak_descriptors
        return res

    def plot(
        self,
        fig_params: Optional[dict] = None,
        line_params: Optional[dict] = None,
        fill_params: Optional[dict] = None,
        palette: Optional[str] = None,
        show: bool = True
    ) -> bokeh.plotting.Figure:     # pragma: no cover
        """
        Plot the chromatogram.

        Parameters
        ----------
        fig_params : dict or None, default=None
            key-value parameters to pass to ``bokeh.plotting.figure``.
        line_params : dict or None, default=None
            key-value parameters to pass to ``bokeh.plotting.Figure.line``.
            Used to draw the chromatogram line.
        fill_params : dict or None, default=None
            key-value parameters to pass to ``bokeh.plotting.Figure.varea``.
            Used to fill the area under the peaks.
        palette : str or None, default=None
            Color palette used to fill the area under the peaks.
        show : bool, default=True
            If True calls ``bokeh.plotting.show`` on the Figure.

        Returns
        -------
        bokeh.plotting.Figure

        """
        default_fig_params = _plot_bokeh.get_chromatogram_figure_params()
        if fig_params:
            default_fig_params.update(fig_params)
            fig_params = default_fig_params
        else:
            fig_params = default_fig_params
        fig = bokeh.plotting.Figure(**fig_params)

        _plot_bokeh.add_line(fig, self.rt, self.spint, line_params=line_params)
        if self.peaks:
            _plot_bokeh.fill_peaks(
                fig,
                self.rt,
                self.spint,
                self.peaks,
                palette,
                varea_params=fill_params
            )
        _plot_bokeh.set_chromatogram_axis_params(fig)
        if show:
            bokeh.plotting.show(fig)
        return fig


class Roi(Chromatogram):
    """
    m/z traces where a chromatographic peak may be found. m/z information
    is stored besides time and intensity information.

    Subclassed from Chromatogram. Used for feature detection in LCMS data.

    Attributes
    ----------
    rt : array
        retention time in each scan.
    spint : array
        intensity in each scan.
    mz : array
        m/z in each scan.
    scan : array
        scan numbers where the ROI is defined.
    mode : {"uplc", "hplc"}
        Analytical platform used separation. Sets default values for peak
        detection.

    """
    def __init__(self, spint: np.ndarray, mz: np.ndarray, rt: np.ndarray,
                 scan: np.ndarray, mode: str = "uplc"):
        super(Roi, self).__init__(rt, spint, mode=mode)
        self.mz = mz
        self.scan = scan

    def fill_nan(self, fill_value: Optional[float] = None):
        """
        Fill missing intensity values using linear interpolation.

        Parameters
        ----------
        fill_value : float or None
            Missing intensity values are replaced with this value. If None,
            values are filled using linear interpolation.

        """

        # if the first or last values are missing, assign an intensity value
        # of zero. This prevents errors in the interpolation and makes peak
        # picking work better.

        if np.isnan(self.spint[0]):
            self.spint[0] = 0
        if np.isnan(self.spint[-1]):
            self.spint[-1] = 0

        missing = np.isnan(self.spint)
        mz_mean = np.nanmean(self.mz)
        if fill_value is None:
            interpolator = interp1d(self.rt[~missing], self.spint[~missing])
            self.mz[missing] = mz_mean
            self.spint[missing] = interpolator(self.rt[missing])
        else:
            self.mz[missing] = mz_mean
            self.spint[missing] = fill_value

    def get_peaks_mz(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Computes the weighted mean of the m/z for each peak and the m/z
        standard deviation.

        Returns
        -------
        mean_mz : array
        mz_std : array

        """
        mz_std = np.zeros(len(self.peaks))
        mz_mean = np.zeros(len(self.peaks))
        for k, peak in enumerate(self.peaks):
            mz_std[k] = self.mz[peak.start:peak.end].std()
            mz_mean[k] = peak.get_loc(self.mz, self.spint)
        return mz_mean, mz_std


def get_lc_filter_peak_params(lc_mode: str) -> dict:
    """
    Default filters for peaks detected in LC data.

    Parameters
    ----------
    lc_mode : {"hplc", "uplc"}
        HPLC assumes typical experimental conditions for HPLC experiments:
        longer columns with particle size greater than 3 micron. UPLC is for
        data acquired with short columns with particle size lower than 3 micron.

    Returns
    -------
    filters : dict
        filters to pass to :py:func:`tidyms.peaks.get_peak_descriptors`.

    """
    if lc_mode == "hplc":
        filters = {"width": (10, 90), "snr": (5, None)}
    elif lc_mode == "uplc":
        filters = {"width": (4, 60), "snr": (5, None)}
    else:
        msg = "`mode` must be `hplc` or `uplc`"
        raise ValueError(msg)
    return filters


def get_find_centroid_params(instrument: str) -> dict:
    """
    Set default parameters to find_centroid method using instrument information.

    Parameters
    ----------
    instrument : {"qtof", "orbitrap"}

    Returns
    -------
    params : dict

    """
    params = {"min_snr": 10}
    if instrument == "qtof":
        md = 0.01
    else:   # orbitrap
        md = 0.005
    params["min_distance"] = md
    return params


class RoiMaker:
    """
    Helper class used by make_roi to create Roi instances from raw data.

    Attributes
    ----------
    mz_mean: numpy.ndarray
        mean value of mz for a given row in mz_array. Used to add new values
        based on a tolerance. Updated after adding a new column
    n_missing: numpy.ndarray
        number of consecutive missing values. Used to detect finished rois
    roi: list[_TemporaryRoi]
    """

    def __init__(self, mz_seed: np.ndarray, max_missing: int = 1,
                 min_length: int = 5, min_intensity: float = 0,
                 tolerance: float = 0.005, multiple_match: str = "closest",
                 mz_reduce: Optional[Callable] = None,
                 sp_reduce: Union[str, Callable] = "sum",
                 targeted: bool = False):
        """

        Parameters
        ----------
        mz_seed: numpy.ndarray
            initial values to build rois
        max_missing: int
            maximum number of missing consecutive values. when a row surpass
            this number the roi is flagged as finished.
        min_length: int
            The minimum length of a finished roi to be considered valid before
            being added to the roi list.
        min_intensity: float
        tolerance: float
            mz tolerance used to connect values.
        multiple_match: {"closest", "reduce"}
            how to match peaks when there is more than one match. If mode is
            `closest`, then the closest peak is assigned as a match and the
            others are assigned to no match. If mode is `reduce`, then a unique
            mz and intensity value is generated using the reduce function in
            `mz_reduce` and `spint_reduce` respectively.
        mz_reduce: callable, optional
            function used to reduce mz values. Can be a function accepting
            numpy arrays and returning numbers. Only used when `multiple_match`
            is ``"reduce"``. See the following prototype:

            def mz_reduce(mz_match: np.ndarray) -> float:
                pass

            If None, m/z values are reduced using the mean.

        sp_reduce: str or callable
            function used to reduce spint values. Can be a function accepting
            numpy arrays and returning numbers. Only used when `multiple_match`
            is ``"reduce"``. To use custom functions see the prototype shown on
            `mz_reduce`.
        """
        if multiple_match not in ["closest", "reduce"]:
            msg = "Valid modes are closest or reduce"
            raise ValueError(msg)

        if mz_reduce == "mean":
            self._mz_reduce = np.mean
        else:
            self._mz_reduce = mz_reduce

        if sp_reduce == "mean":
            self._spint_reduce = np.mean
        elif sp_reduce == "sum":
            self._spint_reduce = np.sum
        else:
            self._spint_reduce = sp_reduce

        # temporary roi data
        self.mz_mean = np.unique(mz_seed.copy())
        # roi index maps the values in mz_mean to a temp roi in temp_roi_dict
        self.roi_index = np.arange(mz_seed.size)
        self.n_missing = np.zeros_like(mz_seed, dtype=int)
        self.max_intensity = np.zeros_like(mz_seed)
        self.length = np.zeros_like(mz_seed, dtype=int)
        self.temp_roi_dict = {x: _make_temporary_roi() for x in self.roi_index}
        self.roi = list()

        # parameters used to build roi
        self.min_intensity = min_intensity
        self.max_missing = max_missing
        self.min_length = min_length
        self.tolerance = tolerance
        self.multiple_match = multiple_match
        self.targeted = targeted

    def extend_roi(self, mz: np.ndarray, sp: np.ndarray, scan: int):
        """
        connects mz values with self.mz_mean to extend existing roi.
        Non-matching mz values are used to create new temporary roi.

        """
        self.n_missing += 1

        if mz.size:
            # find matching and non-matching mz values
            match_index, mz_match, sp_match, mz_no_match, sp_no_match = \
                _match_mz(
                    self.mz_mean, mz, sp, self.tolerance,
                    self.multiple_match, self._mz_reduce, self._spint_reduce
                )

            # extend matching roi
            for k, k_mz, k_sp in zip(match_index, mz_match, sp_match):
                k_temp_roi = self.temp_roi_dict[self.roi_index[k]]
                _append_to__roi(k_temp_roi, k_mz, k_sp, scan)

            # update mz_mean and missing values
            updated_mean = (
                (
                    self.mz_mean[match_index] * self.length[match_index]
                    + mz_match
                )
                / (self.length[match_index] + 1)
            )

            self.length[match_index] += 1
            # reset missing count for matching roi
            self.n_missing[match_index] = 0
            self.max_intensity[match_index] = \
                np.maximum(self.max_intensity[match_index], sp_match)

            # if there are non-matching mz values, use them to build new rois.
            # in targeted mode, only roi with specified mz values are built
            if not self.targeted:
                self.mz_mean[match_index] = updated_mean
                self.create_new_roi(mz_no_match, sp_no_match, scan)

    def store_completed_roi(self):
        """
        store completed ROIs. Valid ROI are appended toi roi attribute.
        The validity of the ROI is checked based on roi length and minimum
        intensity.

        """

        # check completed rois
        is_completed = self.n_missing > self.max_missing

        # length and intensity check
        is_valid_roi = ((self.length >= self.min_length) &
                        (self.max_intensity >= self.min_intensity))

        # add valid roi to roi list
        completed_index = np.where(is_completed)[0]
        for ind in completed_index:
            roi_ind = self.roi_index[ind]
            finished_roi = self.temp_roi_dict.pop(roi_ind)
            if is_valid_roi[ind]:
                self.roi.append(finished_roi)

        # remove completed roi
        if self.targeted:
            self.n_missing[is_completed] = 0
            self.length[is_completed] = 0
            self.max_intensity[is_completed] = 0
            max_roi_ind = self.roi_index.max()
            n_completed = is_completed.sum()
            new_indices = np.arange(max_roi_ind + 1,
                                    max_roi_ind + 1 + n_completed)
            self.roi_index[is_completed] = new_indices
            new_tmp_roi = {k: _make_temporary_roi() for k in new_indices}
            self.temp_roi_dict.update(new_tmp_roi)
        else:
            self.mz_mean = self.mz_mean[~is_completed]
            self.n_missing = self.n_missing[~is_completed]
            self.length = self.length[~is_completed]
            self.roi_index = self.roi_index[~is_completed]
            self.max_intensity = self.max_intensity[~is_completed]

    def create_new_roi(self, mz: np.ndarray, sp: np.ndarray, scan: int):
        """
        Creates new temporary roi from non-matching values

        """

        # finds roi index for new temp roi and update metadata
        max_index = self.roi_index.max()
        new_indices = np.arange(mz.size) + max_index + 1
        mz_mean_tmp = np.hstack((self.mz_mean, mz))
        roi_index_tmp = np.hstack((self.roi_index, new_indices))
        n_missing_tmp = np.zeros_like(new_indices, dtype=int)
        n_missing_tmp = np.hstack((self.n_missing, n_missing_tmp))
        length_tmp = np.ones_like(new_indices, dtype=int)
        length_tmp = np.hstack((self.length, length_tmp))
        max_int_tmp = np.zeros_like(new_indices, dtype=float)
        max_int_tmp = np.hstack((self.max_intensity, max_int_tmp))

        # temp roi creation
        for k_index, k_mz, k_sp in zip(new_indices, mz, sp):
            new_roi = _TemporaryRoi([k_mz], [k_sp], [scan])
            self.temp_roi_dict[k_index] = new_roi

        # replace new temp roi metadata
        # roi extension is done using bisection search, all values are sorted
        # using the mz values
        sorted_index = np.argsort(mz_mean_tmp)
        self.mz_mean = mz_mean_tmp[sorted_index]
        self.roi_index = roi_index_tmp[sorted_index]
        self.n_missing = n_missing_tmp[sorted_index]
        self.length = length_tmp[sorted_index]
        self.max_intensity = max_int_tmp[sorted_index]

    def flag_as_completed(self):
        self.n_missing[:] = self.max_missing + 1

    def process_completed_roi(
        self,
        valid_scan: List[int],
        rt: np.ndarray,
        pad: int,
        separation: str
    ) -> List[Roi]:
        """
        Converts valid ROI into ROI objects.
        Parameters
        ----------
        valid_scan : list
            Scan numbers used for ROI creation
        rt : array
            rt values associated to each scan
        pad : int
            Number of dummy values to pad the ROI with
        separation : {"uplc", "hplc"}
            separation value to pass to ROI constructor function.

        Returns
        -------
        List[ROI] : List of completed ROI.

        """
        valid_scan = np.array(valid_scan)
        roi_list = list()
        for r in self.roi:
            # converting to deque makes padding easier
            r = _TemporaryRoi(deque(r.mz), deque(r.sp), deque(r.scan))
            _pad_roi(r, pad, valid_scan)
            r = _build_roi(r, rt, valid_scan, separation)
            roi_list.append(r)
        return roi_list


def _match_mz(mz1: np.ndarray, mz2: np.ndarray, sp2: np.ndarray,
              tolerance: float, mode: str, mz_reduce: Callable,
              sp_reduce: Callable
              ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray,
                         np.ndarray]:
    """
    aux function to add method in _RoiProcessor. Find matched values.

    Parameters
    ----------
    mz1: numpy.ndarray
        _RoiProcessor mz_mean
    mz2: numpy.ndarray
        mz values to match
    sp2: numpy.ndarray
        intensity values associated to mz2
    tolerance: float
        tolerance used to match values
    mode: {"closest", "merge"}
        Behaviour when more than one peak in mz2 matches with a given peak in
        mz1. If mode is `closest`, then the closest peak is assigned as a
        match and the others are assigned to no match. If mode is `merge`, then
        a unique mz and int value is generated using the average of the mz and
        the sum of the intensities.

    Returns
    ------
    match_index: numpy.ndarray
        index when of peaks matching in mz1.
    mz_match: numpy.ndarray
        values of mz2 that matches with mz1
    sp_match: numpy.ndarray
        values of sp2 that matches with mz1
    mz_no_match: numpy.ndarray
    sp_no_match: numpy.ndarray
    """
    closest_index = find_closest(mz1, mz2)
    dmz = np.abs(mz1[closest_index] - mz2)
    match_mask = (dmz <= tolerance)
    no_match_mask = ~match_mask
    match_index = closest_index[match_mask]

    # check multiple_matches
    unique, first_index, count_index = np.unique(match_index,
                                                 return_counts=True,
                                                 return_index=True)

    # set match values
    match_index = unique
    sp_match = sp2[match_mask][first_index]
    mz_match = mz2[match_mask][first_index]

    # compute matches for duplicates
    multiple_match_mask = count_index > 1
    first_index = first_index[multiple_match_mask]
    if first_index.size > 0:
        first_index_index = np.where(count_index > 1)[0]
        count_index = count_index[multiple_match_mask]
        iterator = zip(first_index_index, first_index, count_index)
        if mode == "closest":
            rm_index = list()   # list of duplicate index to remove
            mz_replace = list()
            spint_replace = list()
            for first_ind, index, count in iterator:
                # check which of the duplicate is closest, the rest are removed
                closest = \
                    np.argmin(dmz[match_mask][index:(index + count)]) + index
                mz_replace.append(mz2[match_mask][closest])
                spint_replace.append(sp2[match_mask][closest])
                remove = np.arange(index, index + count)
                remove = np.setdiff1d(remove, closest)
                rm_index.extend(remove)
            # fix rm_index to full mz2 size
            rm_index = np.where(match_mask)[0][rm_index]
            no_match_mask[rm_index] = True
            mz_match[first_index_index] = mz_replace
            sp_match[first_index_index] = spint_replace
        elif mode == "reduce":
            for first_ind, index, count in iterator:
                # check which of the duplicate is closest
                mz_multiple_match = mz2[match_mask][index:(index + count)]
                sp_multiple_match = sp2[match_mask][index:(index + count)]
                mz_match[first_ind] = mz_reduce(mz_multiple_match)
                sp_match[first_ind] = sp_reduce(sp_multiple_match)
        else:
            msg = "mode must be `closest` or `merge`"
            raise ValueError(msg)

    mz_no_match = mz2[no_match_mask]
    sp_no_match = sp2[no_match_mask]
    return match_index, mz_match, sp_match, mz_no_match, sp_no_match


_TemporaryRoi = namedtuple("TemporaryRoi", ["mz", "sp", "scan"])


def _make_temporary_roi() -> _TemporaryRoi:
    return _TemporaryRoi([], [], [])


def _append_to__roi(roi: _TemporaryRoi, mz: float, sp: float,
                    scan: int):
    roi.mz.append(mz)
    roi.sp.append(sp)
    roi.scan.append(scan)


def _pad_roi(roi: _TemporaryRoi, n: int, valid_scan: np.ndarray):
    first_scan = roi.scan[0]
    last_scan = roi.scan[-1]
    start, end = np.searchsorted(valid_scan, [first_scan, last_scan + 1])
    l_pad_index = max(0, start - n)
    nl = start - l_pad_index
    r_pad_index = min(valid_scan.size, end + n)
    nr = r_pad_index - end

    # fill values
    sp_max = max(roi.sp)
    sp_min = min(roi.sp)
    mz_fill = np.mean(roi.mz)
    sp_threshold = 0.75 * sp_max

    # left pad
    sp_fill_left = sp_max if (roi.sp[0] > sp_threshold) else sp_min
    roi.mz.extendleft([mz_fill] * nl)
    roi.sp.extendleft([sp_fill_left] * nl)
    # deque extendleft from right to left
    roi.scan.extendleft(valid_scan[l_pad_index:start][::-1])

    # right pad
    sp_fill_right = sp_max if (roi.sp[-1] > sp_threshold) else sp_min
    roi.mz.extend([mz_fill] * nr)
    roi.sp.extend([sp_fill_right] * nr)
    roi.scan.extend(valid_scan[end:r_pad_index])


def _build_roi(
        roi: _TemporaryRoi, rt: np.ndarray, scans: np.ndarray, separation: str
) -> Roi:
    """
    Converts a TemporaryRoi into a ROI object

    Parameters
    ----------
    roi : TemporaryRoi
    rt: array
        Retention times associated to each scan
    scans : array
        Scans associated used to build the Rois.
    separation : mode to pass to ROI creation.

    Returns
    -------
    Roi

    """
    # build temporal roi arrays, these include scans that must be removed
    # because they are associated with other ms levels.
    first_scan = roi.scan[0]
    last_scan = roi.scan[-1]
    size = last_scan + 1 - first_scan
    mz_tmp = np.ones(size) * np.nan
    spint_tmp = mz_tmp.copy()

    # copy values of the roi to the temporal arrays
    scan_index = np.array(roi.scan) - roi.scan[0]
    mz_tmp[scan_index] = roi.mz
    spint_tmp[scan_index] = roi.sp

    start_ind, end_ind = np.searchsorted(scans,
                                         [first_scan, last_scan + 1])
    scan_tmp = scans[start_ind:end_ind].copy()
    valid_index = scan_tmp - first_scan
    mz_tmp = mz_tmp[valid_index]
    spint_tmp = spint_tmp[valid_index]
    rt_tmp = rt[scan_tmp].copy()

    # temporal sanity check for the roi arrays
    assert rt_tmp.size == mz_tmp.size
    assert rt_tmp.size == spint_tmp.size
    assert rt_tmp.size == scan_tmp.size

    roi = Roi(spint_tmp, mz_tmp, rt_tmp, scan_tmp, mode=separation)
    return roi



import numpy as np
import bokeh.plotting
from bokeh.palettes import all_palettes
from bokeh.models import ColumnDataSource, Segment
from .peaks import Peak
from .utils import SETTINGS
from typing import Dict, Generator, List, Optional

BOKEH_SETTINGS = SETTINGS["bokeh"]


def get_theme_params() -> dict:
    return BOKEH_SETTINGS["theme"]


def get_line_params() -> dict:
    return BOKEH_SETTINGS["line"]


def get_chromatogram_figure_params() -> dict:
    return BOKEH_SETTINGS["chromatogram"]["figure"]


def get_spectrum_figure_params() -> dict:
    return BOKEH_SETTINGS["spectrum"]["figure"]


def get_varea_params() -> dict:
    return BOKEH_SETTINGS["varea"]


def get_palette() -> List[str]:
    palette_params = BOKEH_SETTINGS["palette"]
    return find_palette(**palette_params)


def make_figure(fig_params: Optional[dict]):
    if fig_params is None:
        fig_params = dict()
    return bokeh.plotting.figure(**fig_params)


def find_palette(name: str, size: Optional[int] = None) -> List[str]:
    try:
        palette = bokeh.palettes.all_palettes[name]
        # by default get the palette with the largest size
        if size is None:
            size = max(list(palette.keys()))
        palette = palette[size]
    except KeyError:
        link = "https://docs.bokeh.org/en/latest/docs/reference/palettes.html"
        msg = "Palette not found. Refer to the list of prebuilt palettes at {}"
        raise ValueError(msg.format(link))
    return palette


def palette_cycler(palette: List[str]) -> Generator[str, None, None]:
    ind = 0
    size = len(palette)
    while True:
        yield palette[ind]
        ind = (ind + 1) % size


def add_line(
    figure: bokeh.plotting.Figure,
    x: np.ndarray,
    y: np.ndarray,
    line_params: Optional[dict] = None
):
    """
    Plots a line.

    Parameters
    ----------
    figure : bokeh.plotting.Figure
        key-value parameters to pass into bokeh figure function.
    x : array
    y : array
    line_params : dict
        key-value parameters to pass into bokeh line function.

    """
    default_line_params = get_line_params()
    if line_params:
        default_line_params.update(line_params)
    line_params = default_line_params
    figure.line(x, y, **line_params)


def set_chromatogram_axis_params(fig: bokeh.plotting.Figure):
    xaxis_params = BOKEH_SETTINGS["chromatogram"]["xaxis"]
    yaxis_params = BOKEH_SETTINGS["chromatogram"]["yaxis"]
    fig.xaxis.update(**xaxis_params)
    fig.yaxis.update(**yaxis_params)


def set_ms_spectrum_axis_params(fig: bokeh.plotting.Figure):
    xaxis_params = BOKEH_SETTINGS["spectrum"]["xaxis"]
    yaxis_params = BOKEH_SETTINGS["spectrum"]["yaxis"]
    fig.xaxis.update(**xaxis_params)
    fig.yaxis.update(**yaxis_params)


def fill_peaks(
    figure: bokeh.plotting.Figure,
    x: np.ndarray,
    y: np.ndarray,
    peaks: List[Peak],
    palette: Optional[str] = None,
    varea_params: Optional[Dict] = None,
):
    default_varea_params = get_varea_params()
    if varea_params:
        default_varea_params.update(varea_params)
    varea_params = default_varea_params

    if palette is None:
        palette = get_palette()
    colors = palette_cycler(palette)
    for p in peaks:
        color = next(colors)
        xp = x[p.start:p.end]
        yp = y[p.start:p.end]
        figure.varea(xp, yp, 0, fill_color=color, **varea_params)


def add_stems(
    fig: bokeh.plotting.Figure,
    x: np.ndarray,
    y: np.ndarray,
    line_params: Optional[Dict] = None
):
    default_line_params = get_line_params()
    if line_params:
        default_line_params.update(line_params)
    line_params = default_line_params
    x0 = x
    y0 = np.zeros_like(y)
    source = ColumnDataSource(dict(x0=x0, x1=x, y0=y0, y1=y))
    stems = Segment(x0="x0", x1="x1", y0="y0", y1="y1", **line_params)
    fig.add_glyph(source, stems)



"""
Common routines to work with raw MS data from metabolomics experiments.

Functions
---------
detect_features(path_list) : Perform feature detection on several samples.
feature_correspondence(feature_data) : Match features across different samples
using a combination of clustering algorithms.

"""

import pandas as pd
import numpy as np
from .fileio import MSData
from .container import DataContainer
from .lcms import Roi
from . import validation
from pathlib import Path
from sklearn.cluster import DBSCAN
from sklearn import mixture
from scipy.optimize import linear_sum_assignment
from typing import Optional, Tuple, List, Dict, Union
from IPython.display import clear_output


__all__ = ["detect_features", "feature_correspondence", "make_data_container"]


def detect_features(path: Union[Path, List[str]], separation: str = "uplc",
                    instrument: str = "qtof", roi_params: Optional[dict] = None,
                    smoothing_strength: Optional[float] = 1.0,
                    find_peaks_params: Optional[dict] = None,
                    descriptors: Optional[dict] = None,
                    filters: Optional[dict] = None,
                    verbose: bool = True
                    ) -> Tuple[Dict[str, List[Roi]], pd.DataFrame]:
    """
    Perform feature detection on LC-MS centroid samples.

    Parameters
    ----------
    path: Path or List[str]
        Path can be a list of strings of absolute path representations to mzML
        files in centroid mode or a Path object. Path objects can be used in
        two ways: It can point to a mzML file or to a directory. in the second
        case all mzML files inside the directory will be analyzed.
    separation: {"uplc", "hplc"}
        Analytical platform used for separation. Used to set default the values
        of `detect_peak_params`, `roi_params` and `filter_params`.
    instrument: {"qtof". "orbitrap"}
        MS instrument used for data acquisition. Used to set default value
        of `roi_params`.
    roi_params: dict, optional
        parameters to pass to :py:meth:`tidyms.MSData.make_roi`
    smoothing_strength: positive number, optional
        Width of a gaussian window used to smooth the ROI. If None, no
        smoothing is applied.
    find_peaks_params : dict, optional
        parameters to pass to :py:func:`tidyms.peaks.detect_peaks`
    descriptors : dict, optional
        descriptors to pass to :py:func:`tidyms.peaks.get_peak_descriptors`
    filters : dict, optional
        filters to pass to :py:func:`tidyms.peaks.get_peak_descriptors`
    descriptors : dict, optional
        pass custom descriptors to :py:func:`tidyms.peaks.get_peak_descriptors`
    filters : dict, optional
        pass custom filters to :py:func:`tidyms.peaks.get_peak_descriptors`
    verbose: bool

    Returns
    -------
    roi_dict: dict
        dictionary of sample names to a list of ROI.
    feature_table: DataFrame
        A Pandas DataFrame where each row is a feature detected in a sample and
        each column is a feature descriptor. By default the following
        descriptors are computed:

        mz
            weighted average of the m/z in the peak region.
        mz std
            standard deviation of the m/z in the peak region.
        rt
            weighted average of the retention time in the peak region.
        width
            Chromatographic peak width.
        height
            Height of the chromatographic peak minus the baseline.
        area
            Area of the chromatographic peak. minus the baseline area.
        sample
            The sample name where the feature was detected.

        Also, two additional columns have information to search each feature
        in its correspondent Roi:

        roi_index :
            index in the list of ROI where the feature was detected.
        peak_index :
            index of the peaks attribute of each ROI associated to the feature.

    Notes
    -----
    Features are detected as follows:

    1.  Default parameters are set based on the values of the parameters
        `instrument` and `separation`.
    2.  Regions of interest (ROI) are detected in each sample. See the
        documentation of :py:meth:`tidyms.fileio.MSData.make_roi` for a detailed
        description of how ROI are created from raw data.
    3.  Features (chromatographic peaks) are detected on each ROI. See
        :py:meth:`tidyms.lcms.Chromatogram.find_peaks` for a detailed
        description of how peaks are detected and how descriptors are computed.

    See Also
    --------
    fileio.MSData.make_roi : Finds ROIs in a mzML sample.
    lcms.ROI.find_peaks : Detect peaks and compute peak estimators for a ROI.

    """
    # parameter validation
    # validation.validate_detect_peaks_params(detect_peak_params)
    validation.validate_descriptors(descriptors)
    validation.validate_filters(filters)
    if roi_params is None:
        roi_params = dict()

    path_list = _get_path_list(path)
    roi_dict = dict()
    ft_table_list = list()
    n_samples = len(path_list)
    for k, sample_path in enumerate(path_list):
        sample_name = sample_path.stem
        sample_path_str = str(sample_path)
        ms_data = MSData(sample_path_str, ms_mode="centroid",
                         instrument=instrument, separation=separation)
        k_roi = ms_data.make_roi(**roi_params)

        if verbose:
            clear_output(wait=True)
            msg = "Processing sample {} ({}/{})."
            msg = msg.format(sample_name, k + 1, n_samples)
            print(msg)
            print("Searching features in {} ROI...".format(len(k_roi)), end=" ")

        k_table = _build_feature_table(k_roi,
                                       smoothing_strength=smoothing_strength,
                                       descriptors=descriptors,
                                       filters=filters,
                                       find_peaks_params=find_peaks_params)

        if verbose:
            msg = "Found {} features".format(k_table.shape[0])
            print(msg)
        k_table["sample"] = sample_name
        roi_dict[sample_name] = k_roi
        ft_table_list.append(k_table)
    feature_table = pd.concat(ft_table_list).reset_index(drop=True)
    feature_table["roi index"] = feature_table["roi index"].astype(int)
    feature_table["peak index"] = feature_table["peak index"].astype(int)
    return roi_dict, feature_table


def _get_path_list(path: Union[str, List[str], Path]) -> List[Path]:
    if isinstance(path, str):
        path = Path(path)

    if isinstance(path, list):
        path_list = [Path(x) for x in path]
        for p in path_list:
            # check if all files in the list exists
            if not p.is_file():
                msg = "{} doesn't exist".format(p)
                raise ValueError(msg)
    else:
        if path.is_dir():
            path_list = list(path.glob("*.mzML"))
        elif path.is_file():
            path_list = [path]
        else:
            msg = ("Path must be a string or Path object pointing to a "
                   "directory with mzML files or a list strings with the "
                   "absolute path to mzML")
            raise ValueError(msg)
    return path_list


def _build_feature_table(roi: List[Roi],
                         smoothing_strength: Optional[float] = 1.0,
                         descriptors: Optional[dict] = None,
                         filters: Optional[dict] = None,
                         find_peaks_params: Optional[dict] = None
                         ) -> pd.DataFrame:
    """
    Builds a DataFrame with feature descriptors.

    Parameters
    ----------
    roi : List[Roi]
    find_peaks_params : dict, optional
            parameters to pass to :py:func:`tidyms.peaks.detect_peaks`
    smoothing_strength: positive number, optional
        Width of a gaussian window used to smooth the signal. If None, no
        smoothing is applied.
    descriptors : dict, optional
        descriptors to pass to :py:func:`tidyms.peaks.get_peak_descriptors`
    filters : dict, optional
        filters to pass to :py:func:`tidyms.peaks.get_peak_descriptors`

    Returns
    -------
    DataFrame

    """
    roi_index_list = list()
    peak_index_list = list()
    mz_mean_list = list()
    mz_std_list = list()
    descriptors_list = list()

    for roi_index, k_roi in enumerate(roi):
        k_roi.fill_nan()
        k_params = k_roi.find_peaks(smoothing_strength=smoothing_strength,
                                    descriptors=descriptors,
                                    filters=filters,
                                    find_peaks_params=find_peaks_params)
        n_features = len(k_params)
        descriptors_list.extend(k_params)
        k_mz_mean, k_mz_std = k_roi.get_peaks_mz()
        roi_index_list.append([roi_index] * n_features)
        peak_index_list.append(range(n_features))
        mz_mean_list.append(k_mz_mean)
        mz_std_list.append(k_mz_std)

    roi_index_list = np.hstack(roi_index_list)
    peak_index_list = np.hstack(peak_index_list)
    mz_mean_list = np.hstack(mz_mean_list)
    mz_std_list = np.hstack(mz_std_list)

    ft_table = pd.DataFrame(data=descriptors_list)
    ft_table = ft_table.rename(columns={"loc": "rt"})
    ft_table["mz"] = mz_mean_list
    ft_table["mz std"] = mz_std_list
    ft_table["roi index"] = roi_index_list
    ft_table["peak index"] = peak_index_list
    ft_table = ft_table.dropna(axis=0)
    ft_table["roi index"] = ft_table["roi index"].astype(int)
    ft_table["peak index"] = ft_table["peak index"].astype(int)
    return ft_table


def feature_correspondence(feature_data: pd.DataFrame, mz_tolerance: float,
                           rt_tolerance: float, min_fraction: float = 0.2,
                           min_likelihood: float = 0.0):
    r"""
    Match features across different samples.

    Feature matching is done using the DBSCAN algorithm and Gaussian mixture
    models. After performing feature correspondence, features that come from the
    same species are clustered together.

    Parameters
    ----------
    feature_data: DataFrame
        Feature descriptors obtained from detect_features function
    mz_tolerance: float
        Maximum distance in m/z between two features in a cluster.
    rt_tolerance: float
        Maximum distance in rt between two features in a cluster.
    min_fraction: float
        Minimum fraction of samples forming a cluster.
    min_likelihood: float
        Minimum likelihood required to recover a missing value. Lower values
        will recover more features, but with lower confidence.

    Returns
    -------
    cluster: Series
        The cluster number for each feature.

    See Also
    --------
    detect_features
    make_data_container

    Notes
    -----
    The correspondence algorithm is as follows:

    1.  Features ares clustered using m/z and rt information with the DBSCAN
        algorithm. Because the dispersion in the m/z and r/t dimension is
        independent the Chebyshev distance is used to make the clusters.
        `rt_tolerance` and `mz_tolerance` are used to build the :math:`\epsilon`
        parameter of the model. rt is scaled using these two parameters to have
        the same tolerance in both dimensions in the following way:

        .. math::

            rt_{scaled} = rt * \frac{mz_{tolerance}}{rt_{tolerance}}

        The min_samples parameter is defined using from the minimum_dr
        (minimum detection rate) and the total number of samples in the
        data. This step gives us a matching of the features, but different
        species can be clustered together if they are close, or some features
        can be considered as noise and removed. These cases are analyzed in the
        following steps.
    2.  In this step the possibility that more than one species is present in a
        cluster is explored. The number of species is estimated computing the
        number of repeated features, :math:`n_{repeated}` in the cluster (that
        is, how many features come from only one sample, hoy many come from
        two, etc...). The fraction of repeated samples is computed using the
        total number of samples and then the number of species,
        :math:`n_{species}` is found as the maximum of repetitions with a
        fraction greater than `min_fraction`. Using :math:`n_{species}`, Each
        cluster is fit to a gaussian mixture model. Once again, because
        dispersion in rt and m/z is orthogonal, we used diagonal covariances
        matrices in the GMM. After this step, for each cluster, subclusters
        may be generated if :math:`n_{species}` is greater than one.
    3.  Each subcluster is analyzed then to remove repeated features. To
        remove repeated features the log-likelihood is evaluated for
        features coming from the same sample. The feature with the greatest
        likelihood is kept and the others are flagged as noise.
    4.  The final step is to search for missing features (a feature is
        missing if some of the samples are not present in a subcluster).
        This is done by searching for features that come from missing samples
        in the features flagged as noise by DBSCAN (or in the previous
        step). If a feature from a missing sample has a log-likelihood
        greater than `min_likelihood` then is added to the subcluster. If
        more than one feature is possible the one with the greatest
        likelihood is chosen.

    """
    # sample names are used to search for missing samples.
    sample_names = feature_data["sample"].unique()

    # DBSCAN clustering
    min_samples = int(sample_names.size * min_fraction + 1)
    cluster = _make_initial_cluster(feature_data, mz_tolerance, rt_tolerance,
                                    min_samples)

    # split feature data into clustered data and noise
    data = feature_data[cluster != -1]
    noise = feature_data[cluster == -1]

    # cluster number is converted to a string. This makes easier to assign
    # subclusters using the notation 0-0, 0-1, etc...
    # TODO : maybe there's a better solution to this and it's not necessary
    #   to convert values to str.
    cluster = cluster.astype(str)
    features_per_cluster = _estimate_n_species_per_cluster(data, cluster)

    for name, group in data.groupby(cluster):
        n_ft = features_per_cluster[name]
        # Here each cluster is split into subclusters, repeated values are
        # removed and missing values are searched in noise.
        # Each change is made changing the values in cluster.
        subcluster = _process_cluster(group, noise, cluster, sample_names, name,
                                      n_species=n_ft,
                                      min_likelihood=min_likelihood)
        cluster[subcluster.index] = subcluster

    # map cluster to numbers again
    cluster_value = np.sort(cluster.unique())
    n_cluster = cluster_value.size
    has_noise = "-1" in cluster_value
    # set a feature code for each feature
    if has_noise:
        cluster_names = _make_feature_names(n_cluster - 1)
        cluster_names = ["noise"] + cluster_names
    else:
        cluster_names = _make_feature_names(n_cluster)

    cluster_mapping = dict(zip(cluster_value, cluster_names))
    cluster = cluster.map(cluster_mapping)
    return cluster


def make_data_container(feature_data: pd.DataFrame, cluster: pd.Series,
                        sample_metadata: pd.DataFrame,
                        fill_na: bool = True) -> DataContainer:
    """
    Organizes the detected and matched features into a DataContainer.

    Parameters
    ----------
    feature_data: DataFrame
        DataFrame obtained from detect_features function.
    cluster: pd.Series
        Series obtained from feature_correspondence function.
    sample_metadata: DataFrame
        DataFrame with information from each analyzed sample. The index must
        be the sample names used in feature_data. A column named "class", with
        the class name of each sample is required. For further data processing
        run order information in a column named "order" and analytical batch
        information in a column named "batch" are recommended.
    fill_na: bool, True
        If True fill missing values in the data matrix with zeros.

    Returns
    -------
    DataContainer
    """

    # remove noise
    feature_data["cluster"] = cluster
    not_noise = cluster != "noise"
    feature_data = feature_data[not_noise]

    # compute aggregate statistics for each feature -> feature metadata
    estimators = {"mz": ["mean", "std", "min", "max"],
                  "rt": ["mean", "std", "min", "max"]}
    feature_metadata = feature_data.groupby("cluster").agg(estimators)
    feature_metadata.columns = _flatten_column_multindex(feature_metadata)
    feature_metadata.index.name = "feature"

    # make data matrix
    data_matrix = feature_data.pivot(index="sample", columns="cluster",
                                     values="area")
    data_matrix.columns.name = "feature"
    if fill_na:
        data_matrix = data_matrix.fillna(0)

    # add samples without features as nan rows
    missing_index = sample_metadata.index.difference(data_matrix.index)
    # TODO: manage data inputting
    missing = pd.DataFrame(data=0, index=missing_index,
                           columns=data_matrix.columns)
    data_matrix = pd.concat((data_matrix, missing))
    data_matrix = data_matrix.loc[sample_metadata.index, :]

    dc = DataContainer(data_matrix, feature_metadata, sample_metadata)
    return dc


def _make_feature_names(n_features: int):
    max_ft_str_length = len(str(n_features))

    def ft_formatter(x):
        return "FT" + str(x + 1).rjust(max_ft_str_length, "0")

    ft_names = [ft_formatter(x) for x in range(n_features)]
    return ft_names


def _flatten_column_multindex(df: pd.DataFrame):
    columns = df.columns
    level_0 = columns.get_level_values(0)
    level_1 = columns.get_level_values(1)
    col_name_map = {"mzmean": "mz", "mzstd": "mz std", "mzmin": "mz min",
                    "mzmax": "mz max", "rtmean": "rt", "rtstd": "rt std",
                    "rtmin": "rt min", "rtmax": "rt max"}
    new_names = [col_name_map[x + y] for x, y in zip(level_0, level_1)]
    return new_names


def _make_initial_cluster(feature_data: pd.DataFrame, mz_tolerance: float,
                          rt_tolerance: float, min_samples: int = 8):
    """
    First guess of correspondence between features using DBSCAN algorithm.
    Auxiliary function to feature_correspondence.

    Parameters
    ----------
    feature_data : DataFrame
        DataFrame obtained from `detect_features` function.
    mz_tolerance : float
        Used to build epsilon parameter of DBSCAN
    rt_tolerance : float
        Used to build epsilon parameter of DBSCAN.
    min_samples : int
        parameter to pass to DBSCAN

    Returns
    -------
    cluster : Series
        The assigned cluster by DBSCAN
    """

    ft_points = feature_data.loc[:, ["mz", "rt"]].copy()
    ft_points["rt"] = ft_points["rt"] * mz_tolerance / rt_tolerance
    dbscan = DBSCAN(eps=mz_tolerance, min_samples=min_samples,
                    metric="chebyshev")
    dbscan.fit(ft_points)
    cluster = pd.Series(data=dbscan.labels_, index=feature_data.index)
    return cluster


def _estimate_n_species_per_cluster(df: pd.DataFrame, cluster: pd. Series,
                                    min_dr: float = 0.2):
    """
    Estimates the number of features that forms a cluster.

    The number of features is estimated as follows:
        1. The number of features per sample is counted and normalized
        to the total number of features.
        2. The number of features in a cluster will be the maximum
        normalized number of features per sample greater than the minimum
        detection rate.

    Parameters
    ----------
    df: DataFrame
        Feature data obtained from feature_correspondence function
    min_dr: float, 0.2
        Minimum detection rate.
    """

    # sample_per_cluster counts the number of features that come from the same
    # sample and express it as a fraction of the total number features
    # the number of features in a cluster is the maximum number of samples
    # in a cluster above the minimum detection rate.

    def find_n_cluster(x):
        return x.index[np.where(x > min_dr)[0]][-1]

    sample_per_cluster = (df["sample"].groupby(cluster)
                          .value_counts()
                          .unstack(-1)
                          .fillna(0)
                          .astype(int)
                          .apply(lambda x: x.value_counts(), axis=1)
                          .fillna(0))
    sample_per_cluster = sample_per_cluster / df["sample"].unique().size

    features_per_cluster = sample_per_cluster.apply(find_n_cluster, axis=1)
    return features_per_cluster


def _make_gmm(ft_data: pd.DataFrame, n_feature: int, cluster_name: str):
    """
    fit a gaussian model and set subcluster names for each feature. Auxiliary
    function to process cluster.

    Parameters
    ----------
    ft_data : DataFrame
        The mz and rt columns of the cluster DataFrame
    n_feature : int
        Number of features estimated in the cluster.
    cluster_name: str

    Returns
    -------
    gmm : GaussianMixtureModel fitted with cluster data
    score: The log-likelihood of each feature.
    subcluster : pd.Series with subcluster labels.
    """
    gmm = mixture.GaussianMixture(n_components=n_feature,
                                  covariance_type="diag")
    gmm.fit(ft_data.loc[:, ["mz", "rt"]].values)
    # scores = pd.Series(data=gmm.score_samples(ft_data), index=ft_data.index)
    ft_data["score"] = gmm.score_samples(ft_data.loc[:, ["mz", "rt"]].values)

    # get index of features in the cases where the number of features is greater
    # than the number of components in the gmm
    noise_index = (ft_data
                   .groupby("sample")
                   .filter(lambda x: x.shape[0] > n_feature))

    if not noise_index.empty:
        noise_index = (noise_index
                       .groupby("sample")
                       .apply(lambda x: _noise_ind(x, n_feature))
                       .droplevel(0)
                       .index)
    else:
        noise_index = noise_index.index

    noise = pd.Series(data="-1", index=noise_index)

    # if the number of features is equal to the number of components in the
    # gmm, each feature is assigned to a cluster using the Hungarian algorithm
    # on the posterior probabilities on each component
    subcluster = (ft_data.loc[ft_data.index.difference(noise_index)]
                  .groupby("sample")
                  .filter(lambda x: x.shape[0] <= n_feature)
                  .groupby("sample")
                  .apply(lambda x: _get_best_cluster(x, gmm))
                  .droplevel(0)
                  .astype(str))
    subcluster = subcluster.apply(lambda x: str(cluster_name) + "-" + x)
    subcluster = pd.concat([noise, subcluster])
    subcluster = subcluster.sort_index()
    # TODO: add here the case where n_features < n_components

    # subcluster = pd.Series(data=gmm.predict(ft_data), index=ft_data.index,
    #                        dtype=str)
    scores = 1
    return gmm, scores, subcluster


def _remove_repeated_features(ft_data: pd.DataFrame, subcluster: pd.Series,
                              sample_data: pd.Series, scores: pd.Series):
    """
    Removes repeated samples from a subcluster. If More than one feature comes
    from the same sample, only the sample with the best log-likelihood is
    conserved, the others are flagged with cluster -1. Auxiliary function of
    _process_cluster

    Parameters
    ----------
    ft_data: DataFrame
        The mz and rt columns of the cluster DataFrame
    sample_data: Series
        The sample column of the cluster DataFrame
    scores: Series
        log-likelihood obtained from the GMM.
    subcluster: pd.Series
        subcluster labels. Obtained from make_gmm
    """
    grouped = ft_data.groupby([subcluster, sample_data])
    for _, repeated_ft in grouped:
        n_ft = repeated_ft.shape[0]
        if n_ft > 1:
            ind = repeated_ft.index
            best_ft = [scores[ind].idxmax()]
            ind = ind.difference(best_ft)
            subcluster[ind] = "-1"


def _search_missing_features(cluster: pd.Series, sample_data: pd.Series,
                             n_feature: int, cluster_name: int,
                             sample_names: List[str], noise: pd.DataFrame,
                             subcluster: pd.Series,
                             gmm: mixture.GaussianMixture,
                             min_likelihood: float):
    """
    Search for missing features in noise. Auxiliary function of
    _process_cluster.

    Parameters
    ----------
    sample_data : Series
        The sample column from cluster data
    n_feature : int
    cluster_name : str
    sample_names : list[str]
        The name of all of the samples used.
    noise : DataFrame
    subcluster : Series
    gmm: GaussianMixture
    min_likelihood: float


    Returns
    -------

    """
    # TODO: this function still needs some work
    for k in range(n_feature):
        k = str(k)
        subc_name = str(cluster_name) + "-" + k
        subcluster_samples = sample_data[subcluster == subc_name]
        missing_samples = np.setdiff1d(sample_names, subcluster_samples)

        # TODO: add some kind of filter of mz, rt to reduce time
        # add cluster == "-1" to not consider taken features
        missing_candidates = noise.loc[
            noise["sample"].isin(missing_samples), ["mz", "rt", "sample"]]
        if not missing_candidates.empty:
            candidates_scores = gmm.score_samples(
                missing_candidates.loc[:, ["mz", "rt"]].values)
            is_candidate = candidates_scores > min_likelihood
            is_any_candidate = is_candidate.any()
            if is_any_candidate:
                missing_candidates = missing_candidates.loc[
                    missing_candidates.index[is_candidate]]
                missing_candidates["score"] = candidates_scores[is_candidate]
                candidate_index = missing_candidates.groupby("sample").apply(
                    lambda x: x["score"].idxmax())
                cluster[candidate_index] = subc_name


def _process_cluster(df: pd.DataFrame, noise: pd.DataFrame, cluster: pd.Series,
                     sample_names: list, cluster_name: str,
                     min_likelihood: float, n_species: int):
    """
    Process each cluster obtained from DBSCAN. Auxiliary function to
    `feature_correspondence`.

    Parameters
    ----------
    df : DataFrame
        feature_data values for a given cluster
    noise : DataFrame
        Features flagged as noise by DBSCAN
    cluster : Series
        Cluster values obtained by DBSCAN.
    sample_names : list[str]
        names of the analyzed samples.
    cluster_name : str
        name of the cluster being analyzed
    min_likelihood : float
    n_species: int
        Number of features in the cluster, estimated with
        `estimate_features_per_cluster`.

    Returns
    -------
    subcluster : Series
        The subcluster values.
    """

    ft_data = df.loc[:, ["mz", "rt", "sample"]]
    sample_data = df["sample"]

    if n_species >= 1:
        # fit a Gaussian mixture model using the cluster data
        gmm, scores, subcluster = _make_gmm(ft_data, n_species, cluster_name)
    else:
        subcluster = pd.Series(data="-1", index=df.index)

    # send repeated samples to noise: only the feature with the best
    # score in the subcluster is conserved
    # _remove_repeated_features(ft_data, subcluster, sample_data, scores)

    to_noise = df[subcluster == "-1"]
    if not to_noise.empty:
        noise = pd.concat([noise, to_noise])

    # search missing samples in noise:
    # _search_missing_features(cluster, sample_data, n_species, cluster_name,
    #                          sample_names, noise, subcluster, gmm,
    #                          min_likelihood)
    return subcluster


def _get_best_cluster(x, gmm):
    """
    Assigns a feature to a cluster the posterior probability to each cluster.
    """
    proba = gmm.predict_proba(x.loc[:, ["mz", "rt"]].values)
    rows, cols = proba.shape
    if rows != cols:
        fill = np.zeros(shape=(cols - rows, cols))
        proba = np.vstack((proba, fill))
    _, best_cluster = linear_sum_assignment(proba)
    best_cluster = best_cluster[:rows]
    best_cluster = pd.Series(data=best_cluster, index=x.index)
    return best_cluster


def _noise_ind(x, n):
    """
    search the index of samples that are going to be considered as noise.
    Reduces the number of features from a sample in a cluster until the size is
    equal to n
    """
    ind = x["score"].sort_values().index[:(x.shape[0] - n)]
    return x.loc[ind, :]
